/*
 * MERGE IN VERSION D: NON OVERLAPPING TRIPLES AND NON OVERLAPPING DICTIONARIES
 * copied from C
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <bitset>
#include<dirent.h>
#include <functional>
#include <math.h>
#include "../src/libdcs/CSD_PFC.h"
#include "../src/libdcs/CSD.h"
#include "../src/util/StopWatch.hpp"

#include <HDTManager.hpp>
#include "../src/rdf/RDFParserNtriplesCallback.hpp"
#include "../src/bitsequence/BitSequence375.h"
#include "../src/triples/TriplesList.hpp"
#include "../src/triples/BitmapTriples.hpp"
#include "../src/triples/CompactTriples.hpp"
#include "../src/triples/PlainTriples.hpp"
#include "../src/dictionary/PlainDictionary.hpp"
#include "../src/dictionary/FourSectionDictionary.hpp"
#include "../src/hdt/BasicHDTExtended.hpp"
#include <HDTEnums.hpp>
#include "toolsSplit.hpp"
#include <algorithm>

using namespace std;
using namespace hdt;

void help() {
	cout << "$ mergeHDT_C_toHDT [options]" << endl;
	cout << "\t-i <partitioned_hdt_File>\t\t\tSingle input file" << endl;
	cout << "\t-d\t<inputDirDictionaries>\tDirectoy of the dictionary split"
			<< endl;
	cout << "\t-t\t<inputDirTriples>\tDirectory of the triples split" << endl;
	cout
			<< "\t-v\t<NumberOfVersions>\tWe need the number of versions if input is in several files"
			<< endl;
	cout
			<< "\t-D\t<VersionToDecompress>\tVersion to Decompress, ommit this parameter to decompress all"
			<< endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-o\t<output>\tSave output into output directory." << endl;

	//cout << "\t-v\tVerbose output" << endl;
}

string IntToBin(int number) {
	string result = "";

	do {
		if ((number & 1) == 0)
			result = "0" + result;
		else
			result = "1" + result;

		number >>= 1;
	} while (number);
	cout << "result:" << result << endl;
	reverse(result.begin(), result.end());
	cout << "result:" << result << endl;
	return result;

}

bool comparisonFunc(unsigned char *c1, unsigned char *c2) {
	return strcmp((const char*) c1, (const char*) c2) < 0;
}

int main(int argc, char *argv[]) {

	StopWatch st;

	string query, inputFile, outputDir, inputDirDictionaries, inputDirTriples,
			numOriginalVersionsinGraph, versionToDecompress;
	string outputTimeFile;

	int c;

	IntToBin(8);
	IntToBin(1);
	while ((c = getopt(argc, argv, "hi:o:t:d:v:D:r:")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'i':
			inputFile = optarg;
			break;
		case 'o':
			outputDir = optarg;
			break;
		case 'd':
			inputDirDictionaries = optarg;
			break;
		case 't':
			inputDirTriples = optarg;
			break;
		case 'v':
			numOriginalVersionsinGraph = optarg;
			break;
		case 'D':
			versionToDecompress = optarg;
			break;
		case 'r':
			outputTimeFile = optarg;
			break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	/*if (argc - optind < 1) {
	 cout << "ERROR: You must supply a partitioned File" << endl << endl;
	 help();
	 return 1;
	 }*/

	//inputFile = argv[optind];
	if (inputFile == ""
			&& (inputDirDictionaries == "" || inputDirTriples == "")) {
		cerr
				<< "Incorrect number of parameters! Please provide one input method (-i for single mode output or (-d and -t) for multiple files)"
				<< endl;
		help();
		exit(-1);
	}

	try {

		//prepare the time output
		ostream *outTime;
		ofstream outFTime;

		if (outputTimeFile != "") {
			outFTime.open(outputTimeFile.c_str());
			outTime = &outFTime;
		} else {
			outTime = &cout;
		}

		unsigned int numVersions;
		unsigned int numOriginalVersions;
		unsigned int numCombinationsDictionaries;
		unsigned int numCombinationsTriples;

		//open the file
		istream *in = new ifstream();
		bool singleFileOption = false;
		vector<unsigned int> combinationsDictionariesTemp;
		vector<unsigned int> combinationsTriplesTemp;
		vector<unsigned int> numVersionsTemp;

		if (numOriginalVersionsinGraph == "") {
			exit(-1);
		}
		
		numOriginalVersions = atoi(numOriginalVersionsinGraph.c_str());

		
		for (int i = 0; i < numOriginalVersions; i++) {
			numVersionsTemp.push_back(i + 1); 
		}
		

		if (inputFile != "") {
			singleFileOption = true;
			cout << "Opening " << inputFile << endl;
			((ifstream*) in)->open(inputFile.c_str());
			if (!in->good()) {
				throw "Error opening file to load HDT.";
			}
			in->read((char*) &numVersions, sizeof(unsigned int));
			in->read((char*) &numCombinationsDictionaries,
					sizeof(unsigned int));
			in->read((char*) &numCombinationsTriples, sizeof(unsigned int));
		} else {

			

			unsigned char isFile = 0x8;
			DIR* DirDictionaries;
			DIR* DirTriples;
			struct dirent *DirEntry;
			DirDictionaries = opendir(inputDirDictionaries.c_str());

			while (DirEntry = readdir(DirDictionaries)) {
				if (DirEntry->d_type == isFile) {
					string fileName(DirEntry->d_name);
					fileName = fileName.substr(0, fileName.find("."));
					combinationsDictionariesTemp.push_back(
							atoi(fileName.c_str()));
					
				}
			}
			numCombinationsDictionaries = combinationsDictionariesTemp.size();

			//Read Triples

			DirTriples = opendir(inputDirTriples.c_str());

			while (DirEntry = readdir(DirTriples)) {
				if (DirEntry->d_type == isFile) {
					string fileName(DirEntry->d_name);
					fileName = fileName.substr(0, fileName.find("."));
					combinationsTriplesTemp.push_back(atoi(fileName.c_str()));
					
				}
			}
			numCombinationsTriples = combinationsTriplesTemp.size();

			
		}
		cout << "Read metadata" << endl;

		//Read metadata

		vector<unsigned int> combinationsDictionaries(
				numCombinationsDictionaries);
		vector<unsigned int> combinationsTriples(numCombinationsTriples);

		vector<vector<unsigned int> > combinationsDictionaryperGraph(
				numOriginalVersions);
		vector<vector<unsigned int> > combinationsTriplesperGraph(
				numOriginalVersions);

		// save concrete combinations per dictionary
		cout
				<< "save concrete combinations per dictionary,numCombinationsDictionaries:"
				<< numCombinationsDictionaries << endl;
		for (int i = 0; i < numCombinationsDictionaries; i++) {
			int combination;
			if (singleFileOption) {
				in->read((char*) &combination, sizeof(unsigned int));
				combinationsDictionaries.at(i) = combination;
			} else {
				combination = combinationsDictionariesTemp.at(i);
				combinationsDictionaries.at(i) =
						combinationsDictionariesTemp.at(i);
			}
			//save the graphs associated to this combination (similar to intToBin)
			int numGraph = 0;
			do {
				if ((combination & 1) != 0) {
					combinationsDictionaryperGraph.at(numGraph).push_back(
							combinationsDictionaries.at(i));
				}

				combination >>= 1;
				numGraph++;
			} while (combination);
		}
		cout
				<< "save concrete combinations per triples, numCombinationsTriples:"
				<< numCombinationsTriples << endl;

		// save concrete combinations per triples
		for (int i = 0; i < numCombinationsTriples; i++) {
			int combination;
			if (singleFileOption) {
				in->read((char*) &combination, sizeof(unsigned int));
				combinationsTriples.at(i) = combination;
			} else {
				combination = combinationsTriplesTemp.at(i);
				combinationsTriples.at(i) = combinationsTriplesTemp.at(i);
			}
			//save the graphs associated to this combination (similar to intToBin)
			int numGraph = 0;
			do {
				if ((combination & 1) != 0) {
					combinationsTriplesperGraph.at(numGraph).push_back(
							combinationsTriples.at(i));
				}

				combination >>= 1;
				numGraph++;
			} while (combination);
		}

		

		//*in >> numVersions;
		//*in >> numCombinations;
		cout << "numVersions:" << numOriginalVersions << endl;
		cout << "numCombinationsDictionaries:" << numCombinationsDictionaries
				<< endl;
		cout << "numCombinationsTriples:" << numCombinationsTriples << endl;
		cout << "initalize dicts and triples with positions:"
				<< pow(2, numOriginalVersions) << endl;
		vector<csd::CSD*> dicts_so(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_s(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_p(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_o(pow(2, numOriginalVersions));
		vector<Triples*> triples(pow(2, numOriginalVersions));

		cout << "read Dictionaries of all combinations" << endl;
		int numDictionaries = 0;

		//PlainDictionary* dict = new PlainDictionary(); //dictionary of the new HDT

		for (int i = 0; i < numCombinationsDictionaries; i++) {
			int currentCombination = combinationsDictionaries.at(i);
			istream* tempin = new ifstream();
			if (!singleFileOption) {
				((ifstream*) in)->close();
				istream* tempin = new ifstream();
				stringstream ss;
				ss << inputDirDictionaries << currentCombination << ".dic";
				
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}
			//read dictionaries
			string configFile = "";
			string options = "";
			HDTSpecification spec(configFile);

			

			csd::CSD* dictionaryTemp;
			dicts_so.at(currentCombination) = dictionaryTemp->load(*in);
			dicts_s.at(currentCombination) = dictionaryTemp->load(*in);
			dicts_o.at(currentCombination) = dictionaryTemp->load(*in);
			dicts_p.at(currentCombination) = dictionaryTemp->load(*in);

			//insert into new dictionary
			for (int k = 0; k < dicts_so.at(currentCombination)->getLength();
					k++) {

				string str(
						(char*) dicts_so.at(currentCombination)->extract(
								k + 1));
				
			}
			for (int k = 0; k < dicts_s.at(currentCombination)->getLength();
					k++) {

				string str(
						(char*) dicts_s.at(currentCombination)->extract(k + 1));
				

			}
			for (int k = 0; k < dicts_o.at(currentCombination)->getLength();
					k++) {

				string str(
						(char*) dicts_o.at(currentCombination)->extract(k + 1));
				

			}
			for (int k = 0; k < dicts_p.at(currentCombination)->getLength();
					k++) {

				string str(
						(char*) dicts_p.at(currentCombination)->extract(k + 1));
				

			}

			
			numDictionaries++;

		}

		cout << "numDictionaries:" << numDictionaries << endl;

		cout << "read Triples of all combinations" << endl;
		int numTriples = 0;
		for (int i = 0; i < numCombinationsTriples; i++) {
			int currentCombination = combinationsTriples.at(i);
			istream* tempin = new ifstream();
			if (!singleFileOption) {
				((ifstream*) in)->close();
				istream* tempin = new ifstream();
				stringstream ss;
				ss << inputDirTriples << currentCombination << ".bt";
				
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}
			//read triples
			string configFile = "";
			string options = "";
			HDTSpecification spec(configFile);

			ControlInformation ci_T;
			ci_T.clear();
			ci_T.setType(TRIPLES);
			
			ci_T.load(*in);
			
			triples.at(currentCombination) = new PlainTriples(spec);
			triples.at(currentCombination)->load(*in, ci_T, NULL);
			
			numTriples++;

		}
		cout << "numTriples:" << numTriples << endl;

		

		cout << "ALL LOADED!" << endl;

		//prepare the number of prior elements
		vector<int> totalPriorSO(pow(2, numOriginalVersions), 0);
		vector<int> totalPriorS(pow(2, numOriginalVersions), 0);
		vector<int> totalPriorP(pow(2, numOriginalVersions), 0);
		vector<int> totalPriorO(pow(2, numOriginalVersions), 0);

		//	for (int i = 0; i < numOriginalVersions; i++) {
		for (int v = 0; v < numVersionsTemp.size(); v++) {

			int i = numVersionsTemp.at(v) - 1;

			if (versionToDecompress == ""
					|| (atoi(versionToDecompress.c_str()) == (v + 1))) {

				cout << endl << "-------------" << endl;
				cout << "Version:" << i << endl;
				std::ostream* outGraph;
				ofstream outD;
				std::stringstream outfile;
				outfile << outputDir << (i + 1) << ".output.hdt";
				cout << "outfile:" << outfile.str().c_str() << endl;
				outD.open(outfile.str().c_str());
				outGraph = &outD;

				vector<unsigned int> combinationsGraph =
						combinationsTriplesperGraph.at(i);

				/*
				 * TO MERGE THE HDT I HAVE TO DO A FIRST PASS OVER ALL THE SUBJECT SO SEE WHICH OF THE SO ARE IN FACT SUBJECTS
				 */
				// Mark which of the new SO is a real Subject
				BitSequence375* actualSubjectsofSO = new BitSequence375();

				/*
				 * GET all the combinations of this graph and load the temporal SO dictionary
				 */
				string s = "";
				for (int h = 0; h < numOriginalVersions; h++) {
					s = ((h == v) ? "1" : "0") + s; //just set the version to 1
				}
				int result = std::bitset<32>(s).to_ulong();

				set<int, std::greater<int> > topLevels = getTopLevelsInt(result,
						numOriginalVersions); //order in descending order

				PlainDictionary* dict = new PlainDictionary(); //dictionary of the new HDT
				PlainDictionary* dictTempSO = new PlainDictionary(); //dictionary of the new HDT
				for (set<int>::iterator itLevels = topLevels.begin();
						itLevels != topLevels.end(); ++itLevels) {
					int currlevel = *itLevels;
					
					//insert into new dictionary
					if (dicts_so.at(currlevel) != NULL) {
						
						for (int k = 0; k < dicts_so.at(currlevel)->getLength();
								k++) {
							
							string str(
									(char*) dicts_so.at(currlevel)->extract(
											k + 1));
							
							dictTempSO->insert(str, SUBJECT);
							dictTempSO->insert(str, OBJECT);
							//just insert as objects, and not subjects, defer to the moment in which we know that SO are in fact subjects
							dict->insert(str, OBJECT);
						}
					}
					if (dicts_s.at(currlevel) != NULL) {
						
						for (int k = 0; k < dicts_s.at(currlevel)->getLength();
								k++) {
							
							string str(
									(char*) dicts_s.at(currlevel)->extract(
											k + 1));
							

							dict->insert(str, SUBJECT);
						}
					}
					if (dicts_o.at(currlevel) != NULL) {
						
						for (int k = 0; k < dicts_o.at(currlevel)->getLength();
								k++) {
							
							string str(
									(char*) dicts_o.at(currlevel)->extract(
											k + 1));
							

							dict->insert(str, OBJECT);
						}
					}
					if (dicts_p.at(currlevel) != NULL) {
						
						for (int k = 0; k < dicts_p.at(currlevel)->getLength();
								k++) {
							
							string str(
									(char*) dicts_p.at(currlevel)->extract(
											k + 1));
							

							dict->insert(str, PREDICATE);
						}
					}
				}
				
				if (dicts_so.at(result) != NULL) {
					//get the current one
					for (int k = 0; k < dicts_so.at(result)->getLength(); k++) {

						string str((char*) dicts_so.at(result)->extract(k + 1));

						dictTempSO->insert(str, SUBJECT);
						dictTempSO->insert(str, OBJECT);
						//just insert as objects, and not subjects, defer to the moment in which we know that SO are in fact subjects
						dict->insert(str, OBJECT);
					}
				}
				if (dicts_s.at(result) != NULL) {
					
					for (int k = 0; k < dicts_s.at(result)->getLength(); k++) {
						
						string str((char*) dicts_s.at(result)->extract(k + 1));
						

						dict->insert(str, SUBJECT);
					}
				}
				if (dicts_o.at(result) != NULL) {
					
					for (int k = 0; k < dicts_o.at(result)->getLength(); k++) {
						
						string str((char*) dicts_o.at(result)->extract(k + 1));
						

						dict->insert(str, OBJECT);
					}
				}
				if (dicts_p.at(result) != NULL) {
					
					for (int k = 0; k < dicts_p.at(result)->getLength(); k++) {
						
						string str((char*) dicts_p.at(result)->extract(k + 1));
						
						dict->insert(str, PREDICATE);
					}
				}

				
				dictTempSO->stopProcessing();

				for (int j = 0; j < combinationsGraph.size(); j++) {
					int currentCombination = combinationsGraph.at(j);
				

					//cout << " prepare the number of prior elements" << endl;
					//get the number of top level combinations
					set<int, std::greater<int> > topLevels = getTopLevelsInt(
							currentCombination, numOriginalVersions); //order in descending order
					//std::sort(topLevels.begin(), topLevels.end(), std::greater<int>());

					int numLevels = topLevels.size();

					//cout << "numLevels:" << numLevels << endl;

					vector<int> totalPriorSO(numLevels + 1, 0);
					vector<int> totalPriorS(numLevels + 1, 0);
					vector<int> totalPriorP(numLevels + 1, 0);
					vector<int> totalPriorO(numLevels + 1, 0);
					vector<int> numCombination(numLevels + 1, 0);

					//get the number of SO, S, P and O of these top level combinations

					int prevSO = 0, prevS = 0, prevO = 0, prevP = 0;
					for (set<int>::iterator itLevels = topLevels.begin();
							itLevels != topLevels.end(); ++itLevels) {
						int currlevel = *itLevels;
						
						// check if we have a dictionary in this level
						if (dicts_so.at(currlevel) != NULL) {

							numCombination.push_back(currlevel);
							totalPriorSO.push_back(prevSO);
							prevSO += dicts_so.at(currlevel)->getLength();
							
							totalPriorS.push_back(prevS);
							prevS += dicts_s.at(currlevel)->getLength();
							
							totalPriorP.push_back(prevP);
							prevP += dicts_p.at(currlevel)->getLength();
							
							totalPriorO.push_back(prevO);
							prevO += dicts_o.at(currlevel)->getLength();
							
						}
					}

					//insert current
					
					totalPriorSO.push_back(prevSO);
					totalPriorS.push_back(prevS);
					totalPriorO.push_back(prevO);
					totalPriorP.push_back(prevP);
					numCombination.push_back(currentCombination);

					
					int totalPriorSOInThisCombination = prevSO;
					int totalSInThisCombination = prevS;
					int totalOInThisCombination = prevO;

					if (dicts_so.at(currentCombination) != NULL) {
						totalPriorSOInThisCombination += dicts_so.at(
								currentCombination)->getLength();
						totalSInThisCombination += dicts_s.at(
								currentCombination)->getLength();
						totalOInThisCombination += dicts_o.at(
								currentCombination)->getLength();
					}
					
					// iterate triples
					Triples* currentCT = triples.at(currentCombination);
					IteratorTripleID* it = currentCT->searchAll();

					
					std::vector<int>::iterator low;
					int pos = 0;
					int comb = 0;
					int curSubject = 0, curObject = 0;
					int prevSubject = -1;
					while (it->hasNext()) {
						TripleID tid = *it->next();
						
						unsigned int subject = tid.getSubject();

						if (subject != prevSubject) {

							
							string subjString = "-", predString = "-",
									objString = "-";
							std::vector<int>::iterator low;

							
							if (subject <= totalPriorSOInThisCombination) {
								
								//search which partition holds this ID
								low = std::lower_bound(totalPriorSO.begin(),
										totalPriorSO.end(), subject);
								//std::cout << "lower_bound at position "<< (low - totalPriorSO.begin()) << '\n';
								pos = (low - totalPriorSO.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
								comb = numCombination.at(pos);
								
								subject = subject - totalPriorSO.at(pos);
								

								unsigned char* subjPoint =
										dicts_so.at(comb)->extract(subject);
								subjString = string((const char*) subjPoint);
								

								
								int subjectID = dictTempSO->stringToId(
										subjString, SUBJECT);
								
								actualSubjectsofSO->set((subjectID - 1), true);

								//important to free space
								free(subjPoint);
							}
						}
						prevSubject = subject;
					}
					
				}
			
				//update the final dictionary with the IDs that are really subjects

				IteratorUCharString* it = dictTempSO->getShared();
				uint indexSubj = 0;
				while (it->hasNext()) {
					string subjString((char*) it->next());
					if (actualSubjectsofSO->access(indexSubj)) {
						dict->insert(subjString, SUBJECT);
					}
					indexSubj++;
				}
				dict->stopProcessing();

				// ******** START NEW HDT
				cout << " Prepare new HDT" << endl;
				string configFile;
				string options;
				HDTSpecification spec(configFile);
				spec.setOptions(options);
				// Read Input Files

				BasicHDTExtended newhdt(spec);
				newhdt.startloadPartialDictionaries();

				newhdt.importDictionary(dict);
				//start triples
				newhdt.startloadPartialTriples();

				

				/*
				 * SECOND PASS ACTUALLY INSERTING THE TRIPLES
				 */
				for (int j = 0; j < combinationsGraph.size(); j++) {
					int currentCombination = combinationsGraph.at(j);
					

					/*
					 * prepare the number of prior elements
					 *
					 */

					//cout << " prepare the number of prior elements" << endl;
					//get the number of top level combinations
					set<int, std::greater<int> > topLevels = getTopLevelsInt(
							currentCombination, numOriginalVersions); //order in descending order
					//std::sort(topLevels.begin(), topLevels.end(), std::greater<int>());

					int numLevels = topLevels.size();

					

					vector<int> totalPriorSO(numLevels + 1, 0);
					vector<int> totalPriorS(numLevels + 1, 0);
					vector<int> totalPriorP(numLevels + 1, 0);
					vector<int> totalPriorO(numLevels + 1, 0);
					vector<int> numCombination(numLevels + 1, 0);

					//get the number of SO, S, P and O of these top level combinations

					int prevSO = 0, prevS = 0, prevO = 0, prevP = 0;
					for (set<int>::iterator itLevels = topLevels.begin();
							itLevels != topLevels.end(); ++itLevels) {
						int currlevel = *itLevels;
						
						// check if we have a dictionary in this level
						if (dicts_so.at(currlevel) != NULL) {

							numCombination.push_back(currlevel);
							totalPriorSO.push_back(prevSO);
							prevSO += dicts_so.at(currlevel)->getLength();
							
							totalPriorS.push_back(prevS);
							prevS += dicts_s.at(currlevel)->getLength();
							
							totalPriorP.push_back(prevP);
							prevP += dicts_p.at(currlevel)->getLength();
							
							totalPriorO.push_back(prevO);
							prevO += dicts_o.at(currlevel)->getLength();
							
						}
					}

					//insert current
					
					totalPriorSO.push_back(prevSO);
					totalPriorS.push_back(prevS);
					totalPriorO.push_back(prevO);
					totalPriorP.push_back(prevP);
					numCombination.push_back(currentCombination);

					
					int totalPriorSOInThisCombination = prevSO;
					int totalSInThisCombination = prevS;
					int totalOInThisCombination = prevO;

					if (dicts_so.at(currentCombination) != NULL) {
						totalPriorSOInThisCombination += dicts_so.at(
								currentCombination)->getLength();
						totalSInThisCombination += dicts_s.at(
								currentCombination)->getLength();
						totalOInThisCombination += dicts_o.at(
								currentCombination)->getLength();
					}
					
					// iterate triples
					Triples* currentCT = triples.at(currentCombination);
					IteratorTripleID* it = currentCT->searchAll();

					
					std::vector<int>::iterator low;
					int pos = 0;
					int comb = 0;
					int curSubject = 0, curObject = 0;
					while (it->hasNext()) {
						TripleID tid = *it->next();
						
						unsigned int subject = tid.getSubject();
						unsigned int predicate = tid.getPredicate();
						unsigned int object = tid.getObject();

						unsigned char* subjPoint;
						unsigned char* predPoint;
						unsigned char* objPoint;

						
						string subjString = "-", predString = "-", objString =
								"-";
						std::vector<int>::iterator low;

						if (subject > totalPriorSOInThisCombination) {
							curSubject = subject
									- totalPriorSOInThisCombination;
							
							low = std::lower_bound(totalPriorS.begin(),
									totalPriorS.end(), curSubject);
							
							pos = (low - totalPriorS.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
							comb = numCombination.at(pos);
							
							subject = curSubject - totalPriorS.at(pos);
							

							subjPoint = dicts_s.at(comb)->extract(subject);
							subjString = string((const char*) subjPoint);
							

						} else {
							//search which partition holds this ID

							low = std::lower_bound(totalPriorSO.begin(),
									totalPriorSO.end(), subject);
							
							pos = (low - totalPriorSO.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
							comb = numCombination.at(pos);
							
							subject = subject - totalPriorSO.at(pos);
							

							subjPoint = dicts_so.at(comb)->extract(subject);
							subjString = string((const char*) subjPoint);
							

						}

						//search predicate
						low = std::lower_bound(totalPriorP.begin(),
								totalPriorP.end(), predicate);
						
						pos = (low - totalPriorP.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
						comb = numCombination.at(pos);
						
						predicate = predicate - totalPriorP.at(pos);
						

						predPoint = dicts_p.at(comb)->extract(predicate);
						predString = string((const char*) predPoint);
						

						//search object

						if (object > totalPriorSOInThisCombination) {
							
							curObject = object - totalPriorSOInThisCombination;
							
							low = std::lower_bound(totalPriorO.begin(),
									totalPriorO.end(), curObject);
							
							pos = (low - totalPriorO.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
							comb = numCombination.at(pos);
							
							object = curObject - totalPriorO.at(pos);
							

							objPoint = dicts_o.at(comb)->extract(object);
							objString = string((const char*) objPoint);
							

						} else {
							//search which partition holds this ID

							low = std::lower_bound(totalPriorSO.begin(),
									totalPriorSO.end(), object);
							
							pos = (low - totalPriorSO.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
							comb = numCombination.at(pos);
							
							object = object - totalPriorSO.at(pos);
							

							objPoint = dicts_so.at(comb)->extract(object);
							objString = string((const char*) objPoint);
							

						}

						TripleString tripleString(subjString, predString,
								objString);
						newhdt.processATriple(tripleString);

						// important to free space
						free(predPoint);
						free(subjPoint);
						free(objPoint);
					}

				}

				newhdt.stoploadPartialTriples();
				newhdt.endloadFromPartial("http://example.org");
				cout << "all done, saving HDT" << endl;
				newhdt.saveToHDT(*outGraph, NULL);
				outD.close();
				
			}
		}
		*outTime << st << endl;

		if (outputTimeFile != "") {
			outFTime.close();
		}

	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

