/*
 * QUERY SPLIT IN VERSION B: NON OVERLAPPING TRIPLES AND OVERLAPPING DICTIONARIES
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <bitset>
#include<dirent.h>
#include <math.h>
#include "../src/libdcs/CSD_PFC.h"
#include "../src/libdcs/CSD.h"
#include "../src/util/StopWatch.hpp"
#include "toolsSplit.hpp"

#include <HDTManager.hpp>
#include "../src/rdf/RDFParserNtriplesCallback.hpp"
#include "../src/bitsequence/BitSequence375.h"
#include "../src/triples/TriplesList.hpp"
#include "../src/triples/BitmapTriples.hpp"
#include "../src/dictionary/FourSectionDictionary.hpp"
#include <HDTEnums.hpp>
#include <algorithm>

using namespace std;
using namespace hdt;

void help() {
	cout << "$ querysplitHDT_B [options]" << endl;
	cout << "\t-i <partitioned_hdt_File>\t\t\tSingle input file" << endl;
	cout << "\t-g \t\t\tGenerate Indexes" << endl;
	cout << "\t-d\t<inputDirDictionaries>\tDirectoy of the dictionary split"
			<< endl;
	cout << "\t-t\t<inputDirTriples>\tDirectory of the triples split" << endl;
	cout
			<< "\t-v\t<NumberOfVersions>\tWe need the number of versions if input is in several files"
			<< endl;
	cout
			<< "\t-D\t<VersionToDecompress>\tVersion to Decompress, ommit this parameter to decompress all"
			<< endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-o\t<output>\tSave output into output directory." << endl;
	cout << "\t-r\t<outputTime>\tSave loading&query time to file." << endl;
	cout << "\t-q\t<queryFile>\tFile with SPO queries." << endl;

	//cout << "\t-v\tVerbose output" << endl;
}

string IntToBin(int number) {
	string result = "";

	do {
		if ((number & 1) == 0)
			result = "0" + result;
		else
			result = "1" + result;

		number >>= 1;
	} while (number);
	cout << "result:" << result << endl;
	reverse(result.begin(), result.end());
	cout << "result:" << result << endl;
	return result;

}

bool comparisonFunc(unsigned char *c1, unsigned char *c2) {
	return strcmp((const char*) c1, (const char*) c2) < 0;
}

int main(int argc, char *argv[]) {

	StopWatch st;
	StopWatch stALL;
	StopWatch loadTimer;

	string query, inputFile, outputDir, inputDirDictionaries, inputDirTriples,
			numOriginalVersionsinGraph, versionToDecompress;
	string outputTimeFile;
	string queryFile;

	bool generateIndexes = false;

	int c;

	IntToBin(8);
	IntToBin(1);
	while ((c = getopt(argc, argv, "hi:o:t:d:v:D:r:q:g")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'i':
			inputFile = optarg;
			break;
		case 'g':
			generateIndexes = true;
			break;
		case 'o':
			outputDir = optarg;
			break;
		case 'd':
			inputDirDictionaries = optarg;
			break;
		case 't':
			inputDirTriples = optarg;
			break;
		case 'v':
			numOriginalVersionsinGraph = optarg;
			break;
		case 'D':
			versionToDecompress = optarg;
			break;
		case 'r':
			outputTimeFile = optarg;
			break;
		case 'q':
			queryFile = optarg;
			break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	/*if (argc - optind < 1) {
	 cout << "ERROR: You must supply a partitioned File" << endl << endl;
	 help();
	 return 1;
	 }*/

	//inputFile = argv[optind];
	if (inputFile == ""
			&& (inputDirDictionaries == "" || inputDirTriples == "")) {
		cerr
				<< "Incorrect number of parameters! Please provide one input method (-i for single mode output or (-d and -t) for multiple files)"
				<< endl;
		help();
		exit(-1);
	}

	try {

		//prepare the time output
		ostream *outTime;
		ofstream outFTime;

		if (outputTimeFile != "") {
			outFTime.open(outputTimeFile.c_str());
			outTime = &outFTime;
		} else {
			outTime = &cout;
		}

		unsigned int numVersions;
		unsigned int numOriginalVersions;
		unsigned int numCombinations;

		//open the file
		istream *in = new ifstream();
		bool singleFileOption = false;
		vector<unsigned int> combinationsTemp;
		vector<unsigned int> numVersionsTemp;

		if (numOriginalVersionsinGraph == "") {
			exit(-1);
		}
		
		numOriginalVersions = atoi(numOriginalVersionsinGraph.c_str());

		if (versionToDecompress == "") {
			for (int i = 0; i < numOriginalVersions; i++) {
				numVersionsTemp.push_back(i + 1); 
			}
		} else {
			numVersionsTemp.push_back(atoi(versionToDecompress.c_str()));
		}

		if (inputFile != "") {
			singleFileOption = true;
			cout << "Opening " << inputFile << endl;
			((ifstream*) in)->open(inputFile.c_str());
			if (!in->good()) {
				throw "Error opening file to load HDT.";
			}
			in->read((char*) &numVersions, sizeof(unsigned int));
			in->read((char*) &numCombinations, sizeof(unsigned int));
		} else {

			

			unsigned char isFile = 0x8;
			DIR* DirDictionaries;
			DIR* DirTriples;
			struct dirent *DirEntry;
			DirDictionaries = opendir(inputDirDictionaries.c_str());

			while (DirEntry = readdir(DirDictionaries)) {
				if (DirEntry->d_type == isFile) {
					string fileName(DirEntry->d_name);
					fileName = fileName.substr(0, fileName.find("."));
					combinationsTemp.push_back(atoi(fileName.c_str()));
					cout << "Found a File, combination : " << fileName << endl;
				}
			}
			numCombinations = combinationsTemp.size();

			//Read Triples

			DirTriples = opendir(inputDirTriples.c_str());

			
		}

		//Read metadata

		vector<unsigned int> combinations(numCombinations);
		vector<vector<unsigned int> > combinationsperGraph(numOriginalVersions);

		for (int i = 0; i < numCombinations; i++) {
			int combination;
			if (singleFileOption) {
				in->read((char*) &combination, sizeof(unsigned int));
				combinations.at(i) = combination;
			} else {
				combination = combinationsTemp.at(i);
				combinations.at(i) = combinationsTemp.at(i);
			}
			//save the graphs associated to this combination (similar to intToBin)
			int numGraph = 0;
			do {
				if ((combination & 1) != 0) {
					combinationsperGraph.at(numGraph).push_back(
							combinations.at(i));
				}

				combination >>= 1;
				numGraph++;
			} while (combination);
		}

		

		//*in >> numVersions;
		//*in >> numCombinations;
		cout << "numVersions:" << numOriginalVersions << endl;
		cout << "numCombinations:" << numCombinations << endl;
		cout << "initalize dicts with positions:" << pow(2, numOriginalVersions)
				<< endl;
		vector<Dictionary*> dicts(pow(2, numOriginalVersions));
		vector<Triples*> triples(pow(2, numOriginalVersions));

		cout << "read Dictionaries of all combinations" << endl;
		int numDictionaries = 0;
		for (int i = 0; i < numCombinations; i++) {
			int currentCombination = combinations.at(i);
			istream* tempin = new ifstream();
			if (!singleFileOption) {
				((ifstream*) in)->close();
				istream* tempin = new ifstream();
				stringstream ss;
				ss << inputDirDictionaries << currentCombination << ".dic";
				
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}
			//read dictionaries
			string configFile = "";
			string options = "";
			HDTSpecification spec(configFile);

			ControlInformation ci_D;
			ci_D.clear();
			ci_D.setType(DICTIONARY);
			spec.setOptions(options);
			//cout << "reading Control Information" << endl;
			ci_D.load(*in);
			Dictionary* dictionaryFourSections = new FourSectionDictionary(
					spec);
			//cout << "reading Dictionary" << endl;
			dictionaryFourSections->load(*in, ci_D, NULL);
			//cout << "Update dicts for " << currentCombination<<endl;
			dicts.at(currentCombination) = dictionaryFourSections;
			//cout << "Reading Triples" << endl;
			numDictionaries++;

			//read triples of this combination
			
			if (!singleFileOption) {
				((ifstream*) in)->close();
				stringstream ss;
				ss << inputDirTriples << currentCombination << ".bt";
				
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}

			ControlInformation ci_T;
			ci_T.clear();
			ci_T.setType(TRIPLES);
			
			ci_T.load(*in);
			
			triples.at(currentCombination) = new BitmapTriples(spec);

			triples.at(currentCombination)->load(*in, ci_T, NULL);
			if (generateIndexes)
				triples.at(currentCombination)->generateIndex(NULL);

		}
		cout << "numDictionaries:" << numDictionaries << endl;
		

		cout << "ALL LOADED!" << endl;
		loadTimer.stop();
		cout << "System loaded and ready to query:" << st << endl;

		StopWatch stQueries;
		int numQuery = 0;
		
		for (int v = 0; v < numVersionsTemp.size(); v++) { //we assume we will have 1 for the query
			int i = numVersionsTemp.at(v) - 1;
			cout << "Version:" << i << endl;
			std::ostream* outGraph;
			ofstream outD;
			std::stringstream outfile;
			outfile << outputDir << (i + 1) << ".result.txt";
			
			outD.open(outfile.str().c_str());
			outGraph = &outD;

			std::ifstream inqueryfile(queryFile.c_str());
			std::string query;

			while (std::getline(inqueryfile, query)) {
				numQuery++;
				TripleString searchTriple;
				readQuery(query, searchTriple);
				
				vector<unsigned int> combinationsGraph =
						combinationsperGraph.at(i);

				bool found = false;
				int results = 0;
				for (int j = 0; j < combinationsGraph.size(); j++) {
					int currentCombination = combinationsGraph.at(j);
					
					// iterate triples
					Triples* currentBT = triples.at(currentCombination);
					Dictionary* currentDic = dicts.at(currentCombination);
					IteratorTripleID* it = currentBT->searchAll();

					//search dictionary
					TripleID tid(0, 0, 0);
					int idd1 = 1, idd2 = 1, idd3 = 1;
					if (searchTriple.getSubject() != "?") {
						idd1 = currentDic->stringToId(searchTriple.getSubject(),
								SUBJECT);
						tid.setSubject(idd1);
					}
					if (searchTriple.getPredicate() != "?") {
						idd2 = currentDic->stringToId(
								searchTriple.getPredicate(), PREDICATE);
						tid.setPredicate(idd2);
					}
					if (searchTriple.getObject() != "?") {
						idd3 = currentDic->stringToId(searchTriple.getObject(),
								OBJECT);
						tid.setObject(idd3);
					}
					if (idd1 != 0 && idd2 != 0 && idd3 != 0) {
						

						IteratorTripleID* itSearch = currentBT->search(tid);
						while (itSearch->hasNext()) { //found
							results++;
							TripleID* resultTid = itSearch->next();
							TripleString resultString;
							currentDic->tripleIDtoTripleString(*resultTid,resultString);
						}
					}
					//}

				}
				*outGraph << numQuery << ":";
				
				*outGraph << results << " results";
				*outGraph << endl;
			}
			outD.close();
			
		}
		cout << "Time " << numQuery << " Queries:" << stQueries << endl;
		cout << "Time ALL:" << stALL << endl;



		if (outputTimeFile != "") {
			stQueries.stop();
			stALL.stop();
			*outTime << "Load time seconds: "
					<< ((double) loadTimer.getReal() / 1000000) << endl;
			*outTime << "Query time: "
					<< ((double) stQueries.getReal() / 1000000) << endl;
			*outTime << "Mean Q time: "
					<< ((double) stQueries.getReal() / 1000000 / numQuery)
					<< endl;
			*outTime << "Total processing time: "
					<< ((double) stALL.getReal() / 1000000) << endl;
			outFTime.close();
		}

	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

