/*
 * READ SPLIT IN VERSION D: NON OVERLAPPING TRIPLES AND NON OVERLAPPING DICTIONARIES
 * copied from C
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <bitset>
#include<dirent.h>
#include <functional>
#include <math.h>
#include "../src/libdcs/CSD_PFC.h"
#include "../src/libdcs/CSD.h"
#include "../src/util/StopWatch.hpp"
#include "toolsSplit.hpp"

#include <HDTManager.hpp>
#include "../src/rdf/RDFParserNtriplesCallback.hpp"
#include "../src/bitsequence/BitSequence375.h"
#include "../src/triples/TriplesList.hpp"
#include "../src/triples/BitmapTriples.hpp"
#include "../src/triples/CompactTriples.hpp"
#include "../src/triples/PlainTriples.hpp"
#include "../src/dictionary/FourSectionDictionary.hpp"
#include <HDTEnums.hpp>
#include "toolsSplit.hpp"
#include <algorithm>

using namespace std;
using namespace hdt;

void help() {
	cout << "$ readsplitHDT_C [options]" << endl;
	cout << "\t-i <partitioned_hdt_File>\t\t\tSingle input file" << endl;
	cout << "\t-d\t<inputDirDictionaries>\tDirectoy of the dictionary split"
			<< endl;
	cout << "\t-t\t<inputDirTriples>\tDirectory of the triples split" << endl;
	cout
			<< "\t-v\t<NumberOfVersions>\tWe need the number of versions if input is in several files"
			<< endl;
	cout
			<< "\t-D\t<VersionToDecompress>\tVersion to Decompress, ommit this parameter to decompress all"
			<< endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-o\t<output>\tSave output into output directory." << endl;
	cout << "\t-r\t<outputTime>\tSave loading&query time to file." << endl;
	cout << "\t-q\t<queryFile>\tFile with SPO queries." << endl;

	//cout << "\t-v\tVerbose output" << endl;
}

string IntToBin(int number) {
	string result = "";

	do {
		if ((number & 1) == 0)
			result = "0" + result;
		else
			result = "1" + result;

		number >>= 1;
	} while (number);
	cout << "result:" << result << endl;
	reverse(result.begin(), result.end());
	cout << "result:" << result << endl;
	return result;

}

bool comparisonFunc(unsigned char *c1, unsigned char *c2) {
	return strcmp((const char*) c1, (const char*) c2) < 0;
}

int main(int argc, char *argv[]) {

	StopWatch st;
	StopWatch stALL;
	StopWatch loadTimer;

	string query, inputFile, outputDir, inputDirDictionaries, inputDirTriples,
			numOriginalVersionsinGraph, versionToDecompress;
	string outputTimeFile;
	string queryFile;

	int c;

	IntToBin(8);
	IntToBin(1);
	while ((c = getopt(argc, argv, "hi:o:t:d:v:D:r:q:")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'i':
			inputFile = optarg;
			break;
		case 'o':
			outputDir = optarg;
			break;
		case 'd':
			inputDirDictionaries = optarg;
			break;
		case 't':
			inputDirTriples = optarg;
			break;
		case 'v':
			numOriginalVersionsinGraph = optarg;
			break;
		case 'D':
			versionToDecompress = optarg;
			break;
		case 'r':
			outputTimeFile = optarg;
			break;
		case 'q':
			queryFile = optarg;
			break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	/*if (argc - optind < 1) {
	 cout << "ERROR: You must supply a partitioned File" << endl << endl;
	 help();
	 return 1;
	 }*/

	//inputFile = argv[optind];
	if (inputFile == ""
			&& (inputDirDictionaries == "" || inputDirTriples == "")) {
		cerr
				<< "Incorrect number of parameters! Please provide one input method (-i for single mode output or (-d and -t) for multiple files)"
				<< endl;
		help();
		exit(-1);
	}

	try {

		//prepare the time output
		ostream *outTime;
		ofstream outFTime;

		if (outputTimeFile != "") {
			outFTime.open(outputTimeFile.c_str());
			outTime = &outFTime;
		} else {
			outTime = &cout;
		}

		unsigned int numVersions;
		unsigned int numOriginalVersions;
		unsigned int numCombinationsDictionaries;
		unsigned int numCombinationsTriples;

		//open the file
		istream *in = new ifstream();
		bool singleFileOption = false;
		vector<unsigned int> combinationsDictionariesTemp;
		vector<unsigned int> combinationsTriplesTemp;
		vector<unsigned int> numVersionsTemp;

		if (numOriginalVersionsinGraph == "") {
			exit(-1);
		}
		
		numOriginalVersions = atoi(numOriginalVersionsinGraph.c_str());

		if (versionToDecompress == "") {
			for (int i = 0; i < numOriginalVersions; i++) {
				numVersionsTemp.push_back(i + 1); 
			}
		} else {
			numVersionsTemp.push_back(atoi(versionToDecompress.c_str()));
		}

		if (inputFile != "") {
			singleFileOption = true;
			cout << "Opening " << inputFile << endl;
			((ifstream*) in)->open(inputFile.c_str());
			if (!in->good()) {
				throw "Error opening file to load HDT.";
			}
			in->read((char*) &numVersions, sizeof(unsigned int));
			in->read((char*) &numCombinationsDictionaries,
					sizeof(unsigned int));
			in->read((char*) &numCombinationsTriples, sizeof(unsigned int));
		} else {

			//numVersions = atoi(inputNumVersions.c_str()); //We need the number of versions if input is in several files

			unsigned char isFile = 0x8;
			DIR* DirDictionaries;
			DIR* DirTriples;
			struct dirent *DirEntry;
			DirDictionaries = opendir(inputDirDictionaries.c_str());

			while (DirEntry = readdir(DirDictionaries)) {
				if (DirEntry->d_type == isFile) {
					string fileName(DirEntry->d_name);
					fileName = fileName.substr(0, fileName.find("."));
					combinationsDictionariesTemp.push_back(
							atoi(fileName.c_str()));
					
				}
			}
			numCombinationsDictionaries = combinationsDictionariesTemp.size();

			//Read Triples

			DirTriples = opendir(inputDirTriples.c_str());

			while (DirEntry = readdir(DirTriples)) {
				if (DirEntry->d_type == isFile) {
					string fileName(DirEntry->d_name);
					fileName = fileName.substr(0, fileName.find("."));
					combinationsTriplesTemp.push_back(atoi(fileName.c_str()));
					
				}
			}
			numCombinationsTriples = combinationsTriplesTemp.size();

			
		}
		cout << "Read metadata" << endl;

		//Read metadata

		vector<unsigned int> combinationsDictionaries(
				numCombinationsDictionaries);
		vector<unsigned int> combinationsTriples(numCombinationsTriples);

		vector<vector<unsigned int> > combinationsDictionaryperGraph(
				numOriginalVersions);
		vector<vector<unsigned int> > combinationsTriplesperGraph(
				numOriginalVersions);

		// save concrete combinations per dictionary
		cout
				<< "save concrete combinations per dictionary,numCombinationsDictionaries:"
				<< numCombinationsDictionaries << endl;
		for (int i = 0; i < numCombinationsDictionaries; i++) {
			int combination;
			if (singleFileOption) {
				in->read((char*) &combination, sizeof(unsigned int));
				combinationsDictionaries.at(i) = combination;
			} else {
				combination = combinationsDictionariesTemp.at(i);
				combinationsDictionaries.at(i) =
						combinationsDictionariesTemp.at(i);
			}
			//save the graphs associated to this combination (similar to intToBin)
			int numGraph = 0;
			do {
				if ((combination & 1) != 0) {
					combinationsDictionaryperGraph.at(numGraph).push_back(
							combinationsDictionaries.at(i));
				}

				combination >>= 1;
				numGraph++;
			} while (combination);
		}
		cout
				<< "save concrete combinations per triples, numCombinationsTriples:"
				<< numCombinationsTriples << endl;

		// save concrete combinations per triples
		for (int i = 0; i < numCombinationsTriples; i++) {
			int combination;
			if (singleFileOption) {
				in->read((char*) &combination, sizeof(unsigned int));
				combinationsTriples.at(i) = combination;
			} else {
				combination = combinationsTriplesTemp.at(i);
				combinationsTriples.at(i) = combinationsTriplesTemp.at(i);
			}
			//save the graphs associated to this combination (similar to intToBin)
			int numGraph = 0;
			do {
				if ((combination & 1) != 0) {
					combinationsTriplesperGraph.at(numGraph).push_back(
							combinationsTriples.at(i));
				}

				combination >>= 1;
				numGraph++;
			} while (combination);
		}

		

		//*in >> numVersions;
		//*in >> numCombinations;
		cout << "numVersions:" << numOriginalVersions << endl;
		cout << "numCombinationsDictionaries:" << numCombinationsDictionaries
				<< endl;
		cout << "numCombinationsTriples:" << numCombinationsTriples << endl;
		cout << "initalize dicts and triples with positions:"
				<< pow(2, numOriginalVersions) << endl;
		vector<csd::CSD*> dicts_so(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_s(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_p(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_o(pow(2, numOriginalVersions));
		vector<Triples*> triples(pow(2, numOriginalVersions));

		cout << "read Dictionaries of all combinations" << endl;
		int numDictionaries = 0;
		for (int i = 0; i < numCombinationsDictionaries; i++) {
			int currentCombination = combinationsDictionaries.at(i);
			istream* tempin = new ifstream();
			if (!singleFileOption) {
				((ifstream*) in)->close();
				istream* tempin = new ifstream();
				stringstream ss;
				ss << inputDirDictionaries << currentCombination << ".dic";
				
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}
			//read dictionaries
			string configFile = "";
			string options = "";
			HDTSpecification spec(configFile);

			

			csd::CSD* dictionaryTemp;
			dicts_so.at(currentCombination) = dictionaryTemp->load(*in);
			dicts_s.at(currentCombination) = dictionaryTemp->load(*in);
			dicts_o.at(currentCombination) = dictionaryTemp->load(*in);
			dicts_p.at(currentCombination) = dictionaryTemp->load(*in);

			
			numDictionaries++;

		}
		cout << "numDictionaries:" << numDictionaries << endl;

		cout << "read Triples of all combinations" << endl;
		int numTriples = 0;
		for (int i = 0; i < numCombinationsTriples; i++) {
			int currentCombination = combinationsTriples.at(i);
			istream* tempin = new ifstream();
			if (!singleFileOption) {
				((ifstream*) in)->close();
				istream* tempin = new ifstream();
				stringstream ss;
				ss << inputDirTriples << currentCombination << ".bt";
				
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}
			//read triples
			string configFile = "";
			string options = "";
			HDTSpecification spec(configFile);

			ControlInformation ci_T;
			ci_T.clear();
			ci_T.setType(TRIPLES);
			//cout << "reading Control Information" << endl;
			ci_T.load(*in);
			//cout << "reading COMPACT TRIPLES" << endl;
			//	triples.at(currentCombination) = new CompactTriples(spec);
			triples.at(currentCombination) = new PlainTriples(spec);
			triples.at(currentCombination)->load(*in, ci_T, NULL);
			
			numTriples++;

		}
		cout << "numTriples:" << numTriples << endl;

		

		StopWatch stQueries;
		int numQuery = 0;
		cout << "ALL LOADED!" << endl;
		cout << "System loaded and ready to query:" << st << endl;
		loadTimer.stop();

		//prepare the number of prior elements
		vector<int> totalPriorSO(pow(2, numOriginalVersions), 0);
		vector<int> totalPriorS(pow(2, numOriginalVersions), 0);
		vector<int> totalPriorP(pow(2, numOriginalVersions), 0);
		vector<int> totalPriorO(pow(2, numOriginalVersions), 0);

		
		for (int v = 0; v < numVersionsTemp.size(); v++) { //we assume we will have 1 for the query
			int i = numVersionsTemp.at(v) - 1;
			cout << "Version:" << i << endl;
			std::ostream* outGraph;
			ofstream outD;
			std::stringstream outfile;
			outfile << outputDir << (i + 1) << ".result.txt";
			
			outD.open(outfile.str().c_str());
			outGraph = &outD;

			vector<unsigned int> combinationsGraph =
					combinationsTriplesperGraph.at(i);

			std::ifstream inqueryfile(queryFile.c_str());
			std::string query;

			while (std::getline(inqueryfile, query)) {
				bool found = false;
				
				numQuery++;
				TripleString searchTriple;
				readQuery(query, searchTriple);
				
				vector<unsigned int> combinationsDics =
						combinationsDictionaryperGraph.at(i);

				int foundIDsubject = 0;
				int foundIDobject = 0;
				int foundIDPredicate = 0;
				int combinationFoundSubject = 0;
				int combinationFoundPredicate = 0;
				int combinationFoundObject = 0;
				bool SisSO = false;
				bool OisSO = false;

				//allow other TPs, although they are inefficient with Plain triples
				if (searchTriple.getSubject() == "?") {
					foundIDsubject = -1; // no need to seach for it
					combinationFoundSubject = -1;
				}
				if (searchTriple.getPredicate() == "?") {
					foundIDPredicate = -1; // no need to seach for it
					combinationFoundPredicate = -1;
				}
				if (searchTriple.getObject() == "?") {
					foundIDobject = -1; // no need to seach for it
					combinationFoundObject = -1;
				}

				for (int j = 0; j < combinationsDics.size(); j++) {
					int currentCombination = combinationsDics.at(j);
					

					if (foundIDsubject == 0) {
						unsigned int IDSubject =
								dicts_so.at(currentCombination)->locate(
										(unsigned char*) searchTriple.getSubject().c_str(),
										searchTriple.getSubject().length());
						if (IDSubject != 0) {
							foundIDsubject = IDSubject;
							combinationFoundSubject = currentCombination;
							SisSO = true;
						} else {
							IDSubject =
									dicts_s.at(currentCombination)->locate(
											(unsigned char*) searchTriple.getSubject().c_str(),
											searchTriple.getSubject().length());
							if (IDSubject != 0) {
								foundIDsubject = IDSubject;
								combinationFoundSubject = currentCombination;
							}
						}
					}
					if (foundIDPredicate == 0) {
						unsigned int IDPredicate =
								dicts_p.at(currentCombination)->locate(
										(unsigned char*) searchTriple.getPredicate().c_str(),
										searchTriple.getPredicate().length());
						if (IDPredicate != 0) {
							foundIDPredicate = IDPredicate;
							combinationFoundPredicate = currentCombination;
						}
					}

					if (foundIDobject == 0) {
						unsigned int IDObject =
								dicts_so.at(currentCombination)->locate(
										(unsigned char*) searchTriple.getObject().c_str(),
										searchTriple.getObject().length());
						if (IDObject != 0) {
							foundIDobject = IDObject;
							combinationFoundObject = currentCombination;
							OisSO = true;
						} else {
							IDObject =
									dicts_o.at(currentCombination)->locate(
											(unsigned char*) searchTriple.getObject().c_str(),
											searchTriple.getObject().length());
							if (IDObject != 0) {
								foundIDobject = IDObject;
								combinationFoundObject = currentCombination;
							}
						}
					}

				}

				

				int results = 0;
				if (foundIDsubject != 0 && foundIDPredicate != 0
						&& foundIDobject != 0) {
					//search in the combinations
					
					for (int j = 0; j < combinationsGraph.size(); j++) {
						int currentCombination = combinationsGraph.at(j);
						

						/*
						 * prepare the number of prior elements
						 *
						 */

						
						//get the number of top level combinations
						set<int, std::greater<int> > topLevels =
								getTopLevelsInt(currentCombination,
										numOriginalVersions); //order in descending order
						//std::sort(topLevels.begin(), topLevels.end(), std::greater<int>());

						int numLevels = topLevels.size();

						
						vector<int> totalPriorSO;
						vector<int> totalPriorS;
						vector<int> totalPriorP;
						vector<int> totalPriorO;
						vector<int> numCombination;

						set<int> combinations;

						//get the number of SO, S, P and O of these top level combinations

						int prevSO = 0, prevS = 0, prevO = 0, prevP = 0;
						for (set<int>::iterator itLevels = topLevels.begin();
								itLevels != topLevels.end(); ++itLevels) {
							int currlevel = *itLevels;
							
							// check if we have a dictionary in this level
							if (dicts_so.at(currlevel) != NULL) {

								numCombination.push_back(currlevel);
								combinations.insert(currlevel);
								totalPriorSO.push_back(prevSO);
								prevSO += dicts_so.at(currlevel)->getLength();
								
								totalPriorS.push_back(prevS);
								prevS += dicts_s.at(currlevel)->getLength();
								
								totalPriorP.push_back(prevP);
								prevP += dicts_p.at(currlevel)->getLength();
								
								totalPriorO.push_back(prevO);
								prevO += dicts_o.at(currlevel)->getLength();
								
							}
						}

						//insert current
					
						totalPriorSO.push_back(prevSO);
						totalPriorS.push_back(prevS);
						totalPriorO.push_back(prevO);
						totalPriorP.push_back(prevP);
						numCombination.push_back(currentCombination);
						combinations.insert(currentCombination);

						
						
						int totalPriorSOInThisCombination = prevSO;
						int totalSInThisCombination = prevS;
						int totalOInThisCombination = prevO;

						if (dicts_so.at(currentCombination) != NULL) {
							totalPriorSOInThisCombination += dicts_so.at(
									currentCombination)->getLength();
							totalSInThisCombination += dicts_s.at(
									currentCombination)->getLength();
							totalOInThisCombination += dicts_o.at(
									currentCombination)->getLength();
						}

						//check that subject, predicate and object are in these combinations
						if ((combinations.find(combinationFoundSubject)
								!= combinations.end()
								|| combinationFoundSubject == -1)
								&& (combinations.find(combinationFoundPredicate)
										!= combinations.end()
										|| combinationFoundPredicate == -1)
								&& (combinations.find(combinationFoundObject)
										!= combinations.end()
										|| combinationFoundObject == -1)) {
							std::vector<int>::iterator it;

							//at least the terms are there
							
							//now convert to final ID in triples and search in the Plain Triples

							int levelSubject = 0, levelPredicate = 0,
									levelObject = 0;
							if (combinationFoundSubject != -1) {
								it = find(numCombination.begin(),
										numCombination.end(),
										combinationFoundSubject);
								//int levelSubject = it - numCombination.begin(); //fixme not sure of this
								levelSubject = std::distance(
										numCombination.begin(), it);//fixme not sure of this
								
							}
							if (combinationFoundPredicate != -1) {
								it = find(numCombination.begin(),
										numCombination.end(),
										combinationFoundPredicate);
								levelPredicate = std::distance(
										numCombination.begin(), it); //fixme not sure of this
								
							}
							if (combinationFoundObject != -1) {
								it = find(numCombination.begin(),
										numCombination.end(),
										combinationFoundObject);
								levelObject = std::distance(
										numCombination.begin(), it); //fixme not sure of this
								
							}
							int finalSubjectID = 0, finalPredicateID = 0,
									finalObjectID = 0;

							if (combinationFoundSubject != -1) {
								if (SisSO) {
									finalSubjectID = foundIDsubject
											+ totalPriorSO.at(levelSubject);
								} else {
									finalSubjectID = foundIDsubject
											+ totalPriorSOInThisCombination
											+ totalPriorS.at(levelSubject);
								}
							}
							if (combinationFoundPredicate != -1) {
								finalPredicateID = foundIDPredicate
										+ totalPriorP.at(levelPredicate);
							}
							if (combinationFoundObject != -1) {
								if (OisSO) {
									finalObjectID = foundIDobject
											+ totalPriorSO.at(levelObject);
								} else {
									finalObjectID = foundIDobject
											+ totalPriorSOInThisCombination
											+ totalPriorO.at(levelObject);
								}
							}

							
							// iterate triples
							Triples* currentCT = triples.at(currentCombination);

							//search
							TripleID tid(finalSubjectID, finalPredicateID,
									finalObjectID);
							
							IteratorTripleID* itSearch = currentCT->search(tid);
							while (itSearch->hasNext()) { //found
								results++;

								TripleID* resultTid = itSearch->next();
								/*
								 * CONVERT TO STRING
								 */
								int subject = resultTid->getSubject();
								int object = resultTid->getObject();
								int predicate = resultTid->getPredicate();
								int curSubject = 0, curObject = 0;
								int pos = 0;
								int comb = 0;
								unsigned char* subjPoint;
													unsigned char* predPoint;
													unsigned char* objPoint;
													
													string subjString = "-", predString = "-", objString = "-";
													std::vector<int>::iterator low;

													if (subject > totalPriorSOInThisCombination) {
														curSubject = subject - totalPriorSOInThisCombination;
														
														low = std::lower_bound(totalPriorS.begin(),
																totalPriorS.end(), curSubject);
														//std::cout << "lower_bound at position "<< (low - totalPriorS.begin()) << '\n';
														pos = (low - totalPriorS.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
														comb = numCombination.at(pos);
														
														subject = curSubject - totalPriorS.at(pos);
														
														subjPoint=dicts_s.at(comb)->extract(
																subject);

														subjString = string(
																(const char*) subjPoint);
														

													} else {
														//search which partition holds this ID

														low = std::lower_bound(totalPriorSO.begin(),
																totalPriorSO.end(), subject);
														//std::cout << "lower_bound at position "<< (low - totalPriorSO.begin()) << '\n';
														pos = (low - totalPriorSO.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
														comb = numCombination.at(pos);
														
														subject = subject - totalPriorSO.at(pos);
														

														subjPoint = dicts_so.at(comb)->extract(
																subject);
														subjString = string(
																(const char*) subjPoint);
														//	cout << "subjString SO:" << subjString << endl;

													}

													//search predicate
													low = std::lower_bound(totalPriorP.begin(),
															totalPriorP.end(), predicate);
													//	std::cout << "lower_bound at position "<< (low - totalPriorP.begin()) << '\n';
													pos = (low - totalPriorP.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
													comb = numCombination.at(pos);
													
													predicate = predicate - totalPriorP.at(pos);
													

													 predPoint = dicts_p.at(comb)->extract(predicate);

													predString = string(
															(const char*) predPoint);
													

													//search object

													if (object > totalPriorSOInThisCombination) {
														
														curObject = object - totalPriorSOInThisCombination;
														
														low = std::lower_bound(totalPriorO.begin(),
																totalPriorO.end(), curObject);
														//std::cout << "lower_bound at position "<< (low - totalPriorO.begin()) << '\n';
														pos = (low - totalPriorO.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
														comb = numCombination.at(pos);
														
														object = curObject - totalPriorO.at(pos);
														

														objPoint = dicts_o.at(comb)->extract(
																object);
														objString = string(
																(const char*) objPoint);
														

													} else {
														//search which partition holds this ID

														low = std::lower_bound(totalPriorSO.begin(),
																totalPriorSO.end(), object);
														//std::cout << "lower_bound at position "<< (low - totalPriorSO.begin()) << '\n';
														pos = (low - totalPriorSO.begin()) - 1; //-1 because lower bound returns the first element  which does not compare less than the given value.
														comb = numCombination.at(pos);
														
														object = object - totalPriorSO.at(pos);
														
														objPoint =  dicts_so.at(comb)->extract(
																object);
														objString = string(
																(const char*)objPoint);
														

													}
													
													// important to free space
																		free(predPoint);
																		free(subjPoint);
																		free(objPoint);
							}

							
						}

					}
				}
				*outGraph << numQuery << ":";
				//(found) ? (*outGraph << "TRUE") : (*outGraph << "FALSE");
				*outGraph << results << " results";
				*outGraph << endl;
			}
			outD.close();
			
		}

		cout << "Time " << numQuery << " Queries:" << stQueries << endl;
		cout << "Time ALL:" << stALL << endl;

		if (outputTimeFile != "") {
			stQueries.stop();
			stALL.stop();
			*outTime << "Load time seconds: "
					<< ((double) loadTimer.getReal() / 1000000) << endl;
			*outTime << "Query time: "
					<< ((double) stQueries.getReal() / 1000000) << endl;
			*outTime << "Mean Q time: "
					<< ((double) stQueries.getReal() / 1000000 / numQuery)
					<< endl;
			*outTime << "Total processing time: "
					<< ((double) stALL.getReal() / 1000000) << endl;
			outFTime.close();
		}

	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

