/*
 * READ SPLIT IN VERSION B: NON OVERLAPPING TRIPLES AND OVERLAPPING DICTIONARIES
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <bitset>
#include<dirent.h>
#include <math.h>
#include "../src/libdcs/CSD_PFC.h"
#include "../src/libdcs/CSD.h"
#include "../src/util/StopWatch.hpp"

#include <HDTManager.hpp>
#include "../src/rdf/RDFParserNtriplesCallback.hpp"
#include "../src/bitsequence/BitSequence375.h"
#include "../src/triples/TriplesList.hpp"
#include "../src/triples/BitmapTriples.hpp"
#include "../src/dictionary/FourSectionDictionary.hpp"
#include <HDTEnums.hpp>
#include <algorithm>

using namespace std;
using namespace hdt;

void help() {
	cout << "$ readsplitHDT_B [options]" << endl;
	cout << "\t-i <partitioned_hdt_File>\t\t\tSingle input file" << endl;
	cout << "\t-d\t<inputDirDictionaries>\tDirectoy of the dictionary split"
			<< endl;
	cout << "\t-t\t<inputDirTriples>\tDirectory of the triples split" << endl;
	cout
			<< "\t-v\t<NumberOfVersions>\tWe need the number of versions if input is in several files"
			<< endl;
	cout
			<< "\t-D\t<VersionToDecompress>\tVersion to Decompress, ommit this parameter to decompress all"
			<< endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-o\t<output>\tSave output into output directory." << endl;

	//cout << "\t-v\tVerbose output" << endl;
}

string IntToBin(int number) {
	string result = "";

	do {
		if ((number & 1) == 0)
			result = "0" + result;
		else
			result = "1" + result;

		number >>= 1;
	} while (number);
	cout << "result:" << result << endl;
	reverse(result.begin(), result.end());
	cout << "result:" << result << endl;
	return result;

}

bool comparisonFunc(unsigned char *c1, unsigned char *c2) {
	return strcmp((const char*) c1, (const char*) c2) < 0;
}

int main(int argc, char *argv[]) {

	StopWatch st;

	string query, inputFile, outputDir, inputDirDictionaries, inputDirTriples,
			numOriginalVersionsinGraph, versionToDecompress;
	string outputTimeFile;

	int c;

	IntToBin(8);
	IntToBin(1);
	while ((c = getopt(argc, argv, "hi:o:t:d:v:D:r:")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'i':
			inputFile = optarg;
			break;
		case 'o':
			outputDir = optarg;
			break;
		case 'd':
			inputDirDictionaries = optarg;
			break;
		case 't':
			inputDirTriples = optarg;
			break;
		case 'v':
			numOriginalVersionsinGraph = optarg;
			break;
		case 'D':
			versionToDecompress = optarg;
			break;
		case 'r':
					outputTimeFile = optarg;
					break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	/*if (argc - optind < 1) {
	 cout << "ERROR: You must supply a partitioned File" << endl << endl;
	 help();
	 return 1;
	 }*/

	//inputFile = argv[optind];
	if (inputFile == ""
			&& (inputDirDictionaries == "" || inputDirTriples == "")) {
		cerr
				<< "Incorrect number of parameters! Please provide one input method (-i for single mode output or (-d and -t) for multiple files)"
				<< endl;
		help();
		exit(-1);
	}

	try {

		//prepare the time output
		ostream *outTime;
		ofstream outFTime;

		if(outputTimeFile!="") {
			outFTime.open(outputTimeFile.c_str());
			outTime = &outFTime;
		} else {
			outTime = &cout;
		}

		unsigned int numVersions;
		unsigned int numOriginalVersions;
		unsigned int numCombinations;

		//open the file
		istream *in = new ifstream();
		bool singleFileOption = false;
		vector<unsigned int> combinationsTemp;
		vector<unsigned int> numVersionsTemp;

		if (numOriginalVersionsinGraph == "") {
			exit(-1);
		}
		
		numOriginalVersions = atoi(numOriginalVersionsinGraph.c_str());

		if (versionToDecompress == "") {
			for (int i = 0; i < numOriginalVersions; i++) {
				numVersionsTemp.push_back(i + 1); 
			}
		} else {
			numVersionsTemp.push_back(atoi(versionToDecompress.c_str()));
		}

		if (inputFile != "") {
			singleFileOption = true;
			cout << "Opening " << inputFile << endl;
			((ifstream*) in)->open(inputFile.c_str());
			if (!in->good()) {
				throw "Error opening file to load HDT.";
			}
			in->read((char*) &numVersions, sizeof(unsigned int));
			in->read((char*) &numCombinations, sizeof(unsigned int));
		} else {

			//numVersions = atoi(inputNumVersions.c_str());//We need the number of versions if input is in several files

			unsigned char isFile = 0x8;
			DIR* DirDictionaries;
			DIR* DirTriples;
			struct dirent *DirEntry;
			DirDictionaries = opendir(inputDirDictionaries.c_str());

			while (DirEntry = readdir(DirDictionaries)) {
				if (DirEntry->d_type == isFile) {
					string fileName(DirEntry->d_name);
					fileName = fileName.substr(0, fileName.find("."));
					combinationsTemp.push_back(atoi(fileName.c_str()));
					cout << "Found a File, combination : " << fileName << endl;
				}
			}
			numCombinations = combinationsTemp.size();

			//Read Triples

			DirTriples = opendir(inputDirTriples.c_str());

			
		}

		//Read metadata

		vector<unsigned int> combinations(numCombinations);
		vector<vector<unsigned int> > combinationsperGraph(numOriginalVersions);

		for (int i = 0; i < numCombinations; i++) {
			int combination;
			if (singleFileOption) {
				in->read((char*) &combination, sizeof(unsigned int));
				combinations.at(i) = combination;
			} else {
				combination = combinationsTemp.at(i);
				combinations.at(i) = combinationsTemp.at(i);
			}
			//save the graphs associated to this combination (similar to intToBin)
			int numGraph = 0;
			do {
				if ((combination & 1) != 0) {
					combinationsperGraph.at(numGraph).push_back(
							combinations.at(i));
				}

				combination >>= 1;
				numGraph++;
			} while (combination);
		}

		

		//*in >> numVersions;
		//*in >> numCombinations;
		cout << "numVersions:" << numOriginalVersions << endl;
		cout << "numCombinations:" << numCombinations << endl;
		cout << "initalize dicts with positions:" << pow(2, numOriginalVersions)
				<< endl;
		vector<Dictionary*> dicts(pow(2, numOriginalVersions));
		vector<Triples*> triples(pow(2, numOriginalVersions));

		cout << "read Dictionaries of all combinations" << endl;
		int numDictionaries = 0;
		for (int i = 0; i < numCombinations; i++) {
			int currentCombination = combinations.at(i);
			istream* tempin = new ifstream();
			if (!singleFileOption) {
				((ifstream*) in)->close();
				istream* tempin = new ifstream();
				stringstream ss;
				ss << inputDirDictionaries << currentCombination << ".dic";
				//cout << "SS:" << ss.str() << endl;
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}
			//read dictionaries
			string configFile = "";
			string options = "";
			HDTSpecification spec(configFile);

			ControlInformation ci_D;
			ci_D.clear();
			ci_D.setType(DICTIONARY);
			spec.setOptions(options);
			//cout << "reading Control Information" << endl;
			ci_D.load(*in);
			Dictionary* dictionaryFourSections = new FourSectionDictionary(
					spec);
			//cout << "reading Dictionary" << endl;
			dictionaryFourSections->load(*in, ci_D, NULL);
			//cout << "Update dicts for " << currentCombination<<endl;
			dicts.at(currentCombination) = dictionaryFourSections;
			//cout << "Reading Triples" << endl;
			numDictionaries++;

			//read triples of this combination
			//istream* tempin = new ifstream();
			if (!singleFileOption) {
				//((ifstream*) tempin)->close();
				((ifstream*) in)->close();
				stringstream ss;
				ss << inputDirTriples << currentCombination << ".bt";
				//cout << "SS:" << ss.str() << endl;
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}

			ControlInformation ci_T;
			ci_T.clear();
			ci_T.setType(TRIPLES);
			//	cout << "reading Control Information" << endl;
			ci_T.load(*in);
			//	cout << "reading BT" << endl;
			triples.at(currentCombination) = new BitmapTriples(spec);

			triples.at(currentCombination)->load(*in, ci_T, NULL);

		}
		cout << "numDictionaries:" << numDictionaries << endl;
		

		cout << "ALL LOADED!" << endl;

		
		for (int v = 0; v < numVersionsTemp.size(); v++) {
			int i = numVersionsTemp.at(v) - 1;
			cout << "Version:" << i << endl;
			std::ostream* outGraph;
			ofstream outD;
			std::stringstream outfile;
			outfile << outputDir << (i + 1) << ".output.nt";
			
			outD.open(outfile.str().c_str());
			outGraph = &outD;

			vector<unsigned int> combinationsGraph = combinationsperGraph.at(i);

			for (int j = 0; j < combinationsGraph.size(); j++) {
				int currentCombination = combinationsGraph.at(j);
				
				// iterate triples
				Triples* currentBT = triples.at(currentCombination);
				Dictionary* currentDic = dicts.at(currentCombination);
				IteratorTripleID* it = currentBT->searchAll();

				while (it->hasNext()) {
					TripleID tid = *it->next();
					
					unsigned int subject = tid.getSubject();
					unsigned int predicate = tid.getPredicate();
					unsigned int object = tid.getObject();

					string subjString = currentDic->idToString(subject,
							SUBJECT);
					string predString = currentDic->idToString(predicate,
							PREDICATE);
					string objString = currentDic->idToString(object, OBJECT);

					
					if (subjString.at(0) != '_') {
						*outGraph << "<" << subjString << ">";

					} else {
						*outGraph << subjString;
					}

					//print/return predicate
					*outGraph << " <" << predString << ">";

					if (objString.at(0) == '"') {
						*outGraph << " " << objString;
					} else if (objString.at(0) != '_') {
						*outGraph << " <" << objString << ">";

					} else {
						*outGraph << " " << objString;
					}

					
					*outGraph << " ." << endl;

				}

			}
			outD.close();
			
		}


		*outTime << st << endl;

		if(outputTimeFile!="") {
			outFTime.close();
		}

	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

