/*
 * READ SPLIT IN VERSION C: NON OVERLAPPING DICTIONARIES AND OVERLAPPING TRIPLES
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <bitset>
#include<dirent.h>
#include <math.h>
#include "../src/libdcs/CSD_PFC.h"
#include "../src/libdcs/CSD.h"
#include "../src/util/StopWatch.hpp"

#include <HDTManager.hpp>
#include "../src/rdf/RDFParserNtriplesCallback.hpp"
#include "../src/bitsequence/BitSequence375.h"
#include "../src/triples/TriplesList.hpp"
#include "../src/triples/BitmapTriples.hpp"
#include <HDTEnums.hpp>
#include <algorithm>

using namespace std;
using namespace hdt;

void help() {
	cout << "$ readsplitHDT_C [options]" << endl;
	cout << "\t-i <partitioned_hdt_File>\t\t\tSingle input file" << endl;
	cout << "\t-d\t<inputDirDictionaries>\tDirectoy of the dictionary split"
			<< endl;
	cout << "\t-t\t<inputDirTriples>\tDirectory of the triples split" << endl;
	cout << "\t-v\t<NumberOfVersions>\tWe need the number of versions so far" << endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-o\t<output>\tSave output into output directory." << endl;

	//cout << "\t-v\tVerbose output" << endl;
}

string IntToBin(int number) {
	string result = "";

	do {
		if ((number & 1) == 0)
			result = "0" + result;
		else
			result = "1" + result;

		number >>= 1;
	} while (number);
	cout << "result:" << result << endl;
	reverse(result.begin(), result.end());
	cout << "result:" << result << endl;
	return result;

}

bool comparisonFunc(unsigned char *c1, unsigned char *c2) {
	return strcmp((const char*) c1, (const char*) c2) < 0;
}

int main(int argc, char *argv[]) {

	StopWatch st;

	string query, inputFile, outputDir, inputDirDictionaries, inputDirTriples,numOriginalVersionsinGraph;
	string outputTimeFile;
	int c;

	IntToBin(8);
	IntToBin(1);
	while ((c = getopt(argc, argv, "hi:o:t:d:v:r:")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'i':
			inputFile = optarg;
			break;
		case 'o':
			outputDir = optarg;
			break;
		case 'd':
			inputDirDictionaries = optarg;
			break;
		case 'v':
			numOriginalVersionsinGraph = optarg;
				break;
		case 't':
			inputDirTriples = optarg;
			break;
		case 'r':
					outputTimeFile = optarg;
					break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	/*if (argc - optind < 1) {
	 cout << "ERROR: You must supply a partitioned File" << endl << endl;
	 help();
	 return 1;
	 }*/

	//inputFile = argv[optind];
	if (inputFile == ""
			&& (inputDirDictionaries == "" || inputDirTriples == "")) {
		cerr
				<< "Incorrect number of parameters! Please provide one input method (-i for single mode output or (-d and -t) for multiple files)"
				<< endl;
		help();
		exit(-1);
	}

	try {

		//prepare the time output
		ostream *outTime;
		ofstream outFTime;

		if(outputTimeFile!="") {
			outFTime.open(outputTimeFile.c_str());
			outTime = &outFTime;
		} else {
			outTime = &cout;
		}

		unsigned int numVersions;
		unsigned int numOriginalVersions;
		unsigned int numCombinations;

		//open the file
		istream *in = new ifstream();
		bool singleFileOption = false;
		vector<unsigned int> combinationsTemp;
		vector<unsigned int> numVersionsTemp;

		if (numOriginalVersionsinGraph == "") {
			exit(-1);
		}
		
		numOriginalVersions = atoi(numOriginalVersionsinGraph.c_str());

		if (inputFile != "") {
			singleFileOption = true;
			cout << "Opening " << inputFile << endl;
			((ifstream*) in)->open(inputFile.c_str());
			if (!in->good()) {
				throw "Error opening file to load HDT.";
			}
			in->read((char*) &numVersions, sizeof(unsigned int));
			in->read((char*) &numCombinations, sizeof(unsigned int));

			for (int i = 0; i < numVersions; i++) {
				numVersionsTemp.push_back(i+1); 
			}
		} else {

			unsigned char isFile = 0x8;
			DIR* DirDictionaries;
			DIR* DirTriples;
			struct dirent *DirEntry;
			DirDictionaries = opendir(inputDirDictionaries.c_str());

			while (DirEntry = readdir(DirDictionaries)) {
				if (DirEntry->d_type == isFile) {
					string fileName(DirEntry->d_name);
					fileName = fileName.substr(0, fileName.find("."));
					combinationsTemp.push_back(atoi(fileName.c_str()));
					cout << "Found a File, combination : " << fileName << endl;
				}
			}
			numCombinations = combinationsTemp.size();

			//Read Triples

			DirTriples = opendir(inputDirTriples.c_str());

			while (DirEntry = readdir(DirTriples)) {
				if (DirEntry->d_type == isFile) {
					string fileName(DirEntry->d_name);
					fileName = fileName.substr(0, fileName.find("."));
					numVersionsTemp.push_back(atoi(fileName.c_str()));
					cout << "Found a Triples File, version : " << fileName
							<< endl;
				}
			}
			numVersions = numVersionsTemp.size();
			sort(numVersionsTemp.begin(), numVersionsTemp.end());
		}

		//Read metadata

		vector<unsigned int> combinations(numCombinations);
		vector<vector<unsigned int> > combinationsperGraph(numOriginalVersions);

		for (int i = 0; i < numCombinations; i++) {
			int combination;
			if (singleFileOption) {
				in->read((char*) &combination, sizeof(unsigned int));
				combinations.at(i) = combination;
			} else {
				combination = combinationsTemp.at(i);
				combinations.at(i) = combinationsTemp.at(i);
			}
			//save the graphs associated to this combination (similar to intToBin)
			int numGraph = 0;
			do {
				if ((combination & 1) != 0) {
					combinationsperGraph.at(numGraph).push_back(
							combinations.at(i));
				}

				combination >>= 1;
				numGraph++;
			} while (combination);
		}

		

		//*in >> numVersions;
		//*in >> numCombinations;
		cout << "numVersions:" << numVersions << endl;
		cout << "numCombinations:" << numCombinations << endl;

		vector<BitSequence375*> actualSubjectsofSO(numOriginalVersions); //just marking the real subjects
		vector<BitmapTriples*> bitmapTriplesGraph(numOriginalVersions);

		vector<csd::CSD*> dicts_so(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_s(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_o(pow(2, numOriginalVersions));
		vector<csd::CSD*> dicts_p(pow(2, numOriginalVersions));

		cout << "read Dictionaries of all combinations" << endl;
		int numDictionaries = 0;
		for (int i = 0; i < numCombinations; i++) {
			int currentCombination = combinations.at(i);
			istream* tempin = new ifstream();
			if (!singleFileOption) {
				((ifstream*)in)->close();
				istream* tempin = new ifstream();
				stringstream ss;
				ss << inputDirDictionaries << currentCombination << ".dic";
			
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}
			//read dictionaries

			csd::CSD* dictSO = new csd::CSD_PFC();
			dicts_so.at(currentCombination) = dictSO->load(*in);
			numDictionaries++;
		

			csd::CSD* dictS = new csd::CSD_PFC();
			dicts_s.at(currentCombination) = dictS->load(*in);
			numDictionaries++;
		

			csd::CSD* dictO = new csd::CSD_PFC();
			dicts_o.at(currentCombination) = dictO->load(*in);
			numDictionaries++;
			


			csd::CSD* dictP = new csd::CSD_PFC();
			dicts_p.at(currentCombination) = dictP->load(*in);
			numDictionaries++;
			
		}
		cout << "numDictionaries:" << numDictionaries << endl;
		

		string configFile = "";
		string options = "";
		HDTSpecification spec(configFile);

		spec.setOptions(options);

		cout << "readTriples of all versions" << endl;


		for (int v = 0; v < numVersionsTemp.size(); v++) {
			int i = numVersionsTemp.at(v)-1;

		

			istream* tempin = new ifstream();
			if (!singleFileOption) {
				((ifstream*)tempin)->close();
				stringstream ss;
				ss << inputDirTriples << (i+1) << ".bt";
			
				//get the correct input file in case of multiple files
				((ifstream*) tempin)->open(ss.str().c_str());
				in = tempin;
			}

			cout << "Read version " << i << endl;
			bitmapTriplesGraph.at(i) = new BitmapTriples(spec);
			BitSequence375* bitmap = new BitSequence375();

			ControlInformation ci;
			ci.clear();
			ci.setType(TRIPLES);
		
			ci.load(*in);
		
			bitmapTriplesGraph.at(i)->load(*in, ci, NULL);
	
			actualSubjectsofSO.at(i) = bitmap->load(*in);

		}
		cout << "ALL LOADED!" << endl;

		// LOAD DICTIONARY FOR GRAPHS

		for (int v = 0; v < numVersionsTemp.size(); v++) {
					int i = numVersionsTemp.at(v)-1;
		

			std::ostream* outGraph;
			ofstream outD;
			std::stringstream outfile;
			outfile << outputDir << (i + 1) << ".output.nt";
			cout << "outfile:" << outfile.str().c_str() << endl;
			outD.open(outfile.str().c_str());
			outGraph = &outD;

			std::vector<unsigned char*> so;
			std::vector<unsigned char*> s;
			std::vector<unsigned char*> o;
			std::vector<unsigned char*> p;
			vector<unsigned int> combinationsGraph = combinationsperGraph.at(i);


			for (int j = 0; j < combinationsGraph.size(); j++) {
				int currentCombination = combinationsGraph.at(j);

				

				for (int k = 0;
						k < dicts_so.at(currentCombination)->getLength(); k++) {
					
					so.push_back(
							dicts_so.at(currentCombination)->extract(k + 1));
				}
				
				for (int k = 0; k < dicts_s.at(currentCombination)->getLength();
						k++) {
					
					s.push_back(dicts_s.at(currentCombination)->extract(k + 1));
				}
				
				for (int k = 0; k < dicts_o.at(currentCombination)->getLength();
						k++) {
					
					
					o.push_back(dicts_o.at(currentCombination)->extract(k + 1));
				}
				
				for (int k = 0; k < dicts_p.at(currentCombination)->getLength();
						k++) {
					
					p.push_back(dicts_p.at(currentCombination)->extract(k + 1));
				}
			}
			//sort
			sort(so.begin(), so.end(), comparisonFunc); //comparisonFunc is important because it does not sort alphabetically otherwise
			sort(s.begin(), s.end(), comparisonFunc);
			sort(o.begin(), o.end(), comparisonFunc);
			sort(p.begin(), p.end(), comparisonFunc);

			
			// iterate triples
			BitmapTriples* currentBT = bitmapTriplesGraph.at(i);
			IteratorTripleID* it = currentBT->searchAll();

			//get num Subject is SO
			int numRealSubjectsinSO = actualSubjectsofSO.at(i)->countOnes();
			while (it->hasNext()) {
				TripleID tid = *it->next();
				unsigned int subject = tid.getSubject();
				unsigned int predicate = tid.getPredicate();
				unsigned int object = tid.getObject();
				
				if (subject <= numRealSubjectsinSO) {
					subject = actualSubjectsofSO.at(i)->select1(subject) + 1;
					//print/return subject from so
					//write < > if needed
					if (so.at(subject - 1)[0] != '_') {
						*outGraph << "<" << so.at(subject - 1) << ">";

					} else {
						*outGraph << so.at(subject - 1);
					}
				} else {
					//print/return subject from s
					subject = subject - numRealSubjectsinSO;

					

					
					//write < > if needed
					if (s.at(subject - 1)[0] != '_') {
						*outGraph << "<" << s.at(subject - 1) << ">";
					} else {
						*outGraph << s.at(subject - 1);

					}
				}
				

				//print/return predicate
				*outGraph << " <" << p.at(predicate - 1) << ">";

				if (object <= so.size()) {
					//print return object from so
					
					//write < > if needed
					if (so.at(object - 1)[0] != '_') {
						*outGraph << " <" << so.at(object - 1) << ">";

					} else {
						*outGraph << " " << so.at(object - 1);
					}

				} else {
					//print/return object from o
					
					object = object - so.size();
					
					//write " or < > if needed
					if (o.at(object - 1)[0] == '"') {
						*outGraph << " " << o.at(object - 1);
					} else if (o.at(object - 1)[0] != '_') {
						*outGraph << " <" << o.at(object - 1) << ">";

					} else {
						*outGraph << " " << o.at(object - 1);
					}
				}
				
				*outGraph << " ." << endl;

			}
			outD.close();
			
		}

		*outTime << st << endl;

		if(outputTimeFile!="") {
			outFTime.close();
		}

	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

