/*
 * SPLIT IN VERSION B: NON OVERLAPPING TRIPLES AND OVERLAPPING DICTIONARIES
 */
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <bitset>
#include <math.h>
#include "../src/libdcs/CSD_PFC.h"
#include "../src/libdcs/CSD.h"
#include "../src/util/StopWatch.hpp"

#include <HDTManager.hpp>
#include "../src/rdf/RDFParserNtriplesCallback.hpp"
#include "../src/bitsequence/BitSequence375.h"
#include "../src/triples/TriplesList.hpp"
#include "../src/triples/BitmapTriples.hpp"
#include "../src/dictionary/PlainDictionary.hpp"
#include "../src/dictionary/FourSectionDictionary.hpp"
#include <HDTEnums.hpp>

using namespace std;
using namespace hdt;

class ParseCallback: public RDFCallback {

private:
	long count;
	Dictionary *dictionary;
	BitSequence375* bitmapTriplesGraph;
	IteratorGetPosition* itgetPositionTriple;

public:
	ParseCallback(Dictionary *dictionary, BitSequence375* bitmapTriplesGraph,
			IteratorGetPosition* itgetPositionTriple) :
			dictionary(dictionary), count(0), bitmapTriplesGraph(
					bitmapTriplesGraph), itgetPositionTriple(
					itgetPositionTriple) {
	}
	;
	void processTriple(hdt::TripleString& triple, unsigned long long pos);
};

void help() {
	cout
			<< "$ splitHDT_B [options] <full_hdt_File> <number_of_partitions> <directory_With_Partitions>"
			<< endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-s\t\t\tSave current triples and exit" << endl;
	cout << "\t-o\t<outputSingleFile>\tSave split into one single file ."
			<< endl;
	cout
			<< "\t-d\t<outputDirDictionaries>\tSave dictionary split into several files in the outputDirDictionaries directory."
			<< endl;
	cout
			<< "\t-t\t<outputDirTriples>\tSave triples split into several files in the outputDirTriples directory."
			<< endl;
	cout << "\t-r\t<timeFile>\tSave the performance time to a file."
					<< endl;
	//cout << "\t-v\tVerbose output" << endl;
}

class PositionIteratorUCharString: public IteratorUCharString {
private:
	std::vector<int> vector;
	Dictionary* dictionary;
	unsigned int pos;
public:
	PositionIteratorUCharString(std::vector<int> vector, Dictionary* dictionary) :
			pos(0), dictionary(dictionary), vector(vector) {
	}
	//PositionIteratorUCharString(std::vector<int> vector, Dictionary* dictionary) : vector(vector), dictionary(dictionary) pos(0) { }
	virtual ~PositionIteratorUCharString() {
	}

	virtual bool hasNext() {
		return pos < vector.size();
	}

	virtual unsigned char *next() {
		int id = vector[pos++];
		string ret = "";
		if (id <= dictionary->getNshared()) {
			ret = dictionary->idToString(id, SUBJECT);
		} else if (id <= dictionary->getNsubjects()) {
			ret = dictionary->idToString(id, SUBJECT);
		} else if (id
				> (dictionary->getNumberOfElements()
						- dictionary->getNpredicates())) {
			//	cout<<"id:"<<id<<endl;
			//	cout<<" dictionary->getNumberOfElements():"<< dictionary->getNumberOfElements()<<endl;
			//	cout << "id-(dictionary->getNumberOfElements()- dictionary->getNpredicates()):"
			//			<< (id-(dictionary->getNumberOfElements()
			//				- dictionary->getNpredicates()))<<endl;
			ret = dictionary->idToString(
					(id
							- (dictionary->getNumberOfElements()
									- dictionary->getNpredicates())),
					PREDICATE);
			//	cout<<"ret:"<<ret<<endl;
		} else {
			ret = dictionary->idToString(
					id - dictionary->getNsubjects() + dictionary->getNshared(),
					OBJECT);
		}
		//cerr<<ret<<endl;
		// fflush(stderr);
		/*if (ret.length()==0){
		 cerr<<"ret is 0 with id="<<id<<endl;
		 exit(-1);
		 }*/
		char* classVariableCharPointer = new char[strlen((char*) ret.c_str())
				+ 1];

		// strncpy((char*)(text+bytes), (char*)classVariableCharPointer, currentLength);
		strcpy(classVariableCharPointer, (const char*) ret.c_str());
		return (unsigned char*) classVariableCharPointer;
		//return (unsigned char *) ret.c_str();
	}
};

IteratorGetPosition::IteratorGetPosition(BitmapTriples *trip) :
		triples(trip), adjY(trip->arrayY, trip->bitmapY), adjZ(trip->arrayZ,
				trip->bitmapZ) {

	minY = minZ = 0;
	patX = patY = patZ = 0;
	x = y = z = maxY = maxZ = posZ = posY = 0;

}
int IteratorGetPosition::getNumberofTriple(TripleID &pat) {

	unsigned int retID = 0;
	// assume the order of trip and pat is SPO
	unsigned int patX = pat.getSubject();
	unsigned int patY = pat.getPredicate();
	unsigned int patZ = pat.getObject();

	if (patX != 0 && patY != 0 && patZ != 0) {
		// S P O
		try {
			minY = adjY.find(patX - 1, patY);
			maxY = minY + 1;

			minZ = adjZ.find(minY, patZ);
			retID = minZ;
			maxZ = minZ + 1;

		} catch (const char *ex) {
			// Item not found in list, no results.
			minY = minZ = maxY = maxZ = 0;
		}

		x = patX;
	}
	return retID;
}

TripleID* IteratorGetPosition::getTripleFromPosition(unsigned int position) {

	posY = adjZ.findListIndex(position);

	z = adjZ.get(position);
	y = adjY.get(posY);
	x = adjY.findListIndex(posY) + 1;

	return new TripleID(x, y, z);

}

void testPositionTriples(HDT* hdt) {
	//test triples numbers
	IteratorGetPosition* itPositionTriple = new IteratorGetPosition(
			(BitmapTriples*) (hdt->getTriples()));
	IteratorTripleID* itTriples = hdt->getTriples()->searchAll();
	int numTriples = hdt->getTriples()->getNumberOfElements();
	while (itTriples->hasNext()) {
		TripleID triple = *itTriples->next();
		int position = itPositionTriple->getNumberofTriple(triple);
		cout << triple << ", position:" << position << endl;
	}
	for (int tn = 0; tn < numTriples; tn++) {
		TripleID tid = *itPositionTriple->getTripleFromPosition(tn);
		cout << "position:" << tn << ", triple:" << tid << endl;
	}
	exit(-1);
}

int main(int argc, char *argv[]) {

	StopWatch st;

	string query, inputFile, outputsingleFile, outputDirDictionaries,
			outputDirTriples;
	string outputTimeFile;
	int c;
	bool savecurrentTriple = false;
	while ((c = getopt(argc, argv, "ho:d:t:sr:")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'o':
			outputsingleFile = optarg;
			break;

		case 's':
			savecurrentTriple = true;
			break;
		case 'd':
			outputDirDictionaries = optarg;
			break;
		case 't':
			outputDirTriples = optarg;
			break;
		case 'r':
					outputTimeFile = optarg;
					break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	if (argc - optind < 1) {
		cout << "ERROR: You must supply an HDT File" << endl << endl;
		help();
		return 1;
	}

	inputFile = argv[optind];

	if (argc < 4) {
		cerr << "Incorrect number of parameters!" << endl;
		help();
		exit(0);
	}

	try {

		//prepare the time output
		ostream *outTime;
		ofstream outFTime;

		if(outputTimeFile!="") {
			outFTime.open(outputTimeFile.c_str());
			outTime = &outFTime;
		} else {
			outTime = &cout;
		}

		// Load HDT main file
		HDT *hdt = HDTManager::mapHDT(inputFile.c_str());
		cout << "Opening " << inputFile << endl;
		Dictionary *dictionary = hdt->getDictionary();
		Triples *triples = hdt->getTriples();
		// iterator for the position
		IteratorGetPosition* itPositionTriple = new IteratorGetPosition(
				(BitmapTriples*) (hdt->getTriples()));

		//test triples numbers
		//testPositionTriples(hdt);
		if (savecurrentTriple) {
			//save triples just to compare the size
			std::ostream* outTriples;
			ofstream outT;
			outT.open("OutTriples.bt");
			outTriples = &outT;
			triples->save(*outTriples, *new ControlInformation(), NULL);
			exit(0);
		}
		int numVersions = atoi(argv[optind + 1]);

		vector<BitSequence375*> bitmapsTriplesGraph(numVersions);
		
		vector<BitSequence375*> bitmapsSubjetsGraph(numVersions); //just marking the real subjects

		// Process
		string configFile = "";
		string options = "";
		HDTSpecification spec(configFile);

		spec.setOptions(options);
		for (int i = 0; i < numVersions; i++) {
			bitmapsTriplesGraph.at(i) = new BitSequence375();
			bitmapsSubjetsGraph.at(i) = new BitSequence375();

		}

		cout << "Total elements:" << dictionary->getNumberOfElements() << endl;

		RDFParserNtriplesCallback* parser = new RDFParserNtriplesCallback();

		RDFNotation notation = NTRIPLES;

		string dirPartitions = argv[optind + 2];

		/*
		 * FIRST PASS, ADD THE 1'S TO THE BITMAPS
		 */

		for (int i = 1; i < numVersions + 1; i++) {

			ParseCallback *callback = new ParseCallback(dictionary,
					bitmapsTriplesGraph.at(i - 1), itPositionTriple);
			stringstream ss;
			ss << i;
			string filename = dirPartitions + ss.str() + ".nt";
			cout << filename << endl;
			parser->doParse((const char*) filename.c_str(), "", notation,
					callback);
		}
		delete parser; // Remember to delete iterator to avoid memory leaks!

		/*
		 * COMPUTE THE COMBINATIONS
		 */
		vector<unsigned int> combinations;
		set<unsigned int> combinationsUniq;

		vector<vector<int> > triples_positionsPerCombination(
				pow(2, numVersions));
		vector<ModifiableTriples*> triplesListsperCombination(
				pow(2, numVersions));

		//initialize triplesListsperCombination;
		for (int i = 0; i < pow(2, numVersions); i++) {
			triplesListsperCombination.at(i) = new TriplesList(spec);
		}
		for (int i = 0; i < triples->getNumberOfElements(); i++) {
			string s = "";
			for (int j = 0; j < numVersions; j++) {
				s = (bitmapsTriplesGraph.at(j)->access(i) ? "1" : "0") + s;
			}
			int result = std::bitset<32>(s).to_ulong();
			/*if (result==0){
			 cout<<"found 0 in position:"<<i<<endl;
			 exit(-1);
			 }*/
			//	cout<<"s is:"<<s<<endl;
			//	cout<<"result is:"<<result<<endl;
			combinations.push_back(result);
			combinationsUniq.insert(result);
			//store the position in the corresponding vector

			triples_positionsPerCombination.at(result).push_back(i);
		}

		std::set<unsigned int>::iterator it;
		cout << "* " << combinationsUniq.size() << " combinations *" << endl;
		ostream *out;
		ofstream outF;

		bool singleFileOption = false;
		if (outputsingleFile != "") {
			singleFileOption = true;
			outF.open(outputsingleFile.c_str());
			out = &outF;
		} else {
			out = &cout;
			if (outputDirDictionaries == "" || outputDirTriples == "") {
				cerr
						<< "Please provide one output method (-o for single mode output or (-d and -t) for multiple files)"
						<< endl;
				help();
				exit(-1);
			}
		}
		//singleFileOption:concatenate all dictionaries&triples in a file

		// CREATE SUBDICTIONARIES FOR EACH COMBINATION

		//metadata: save the number of versions and the number of combinations

		//	*out<<(unsigned int)numVersions;
		//	*out<<(unsigned int)combinationsUniq.size();

		if (singleFileOption) {
			out->write((char*) &numVersions, sizeof(unsigned int));
			unsigned int numCombinations = combinationsUniq.size();
			out->write((char*) &numCombinations, sizeof(unsigned int));
			// write combinations
			for (it = combinationsUniq.begin(); it != combinationsUniq.end();
					++it) {
				unsigned int combination = *it;
				out->write((char*) &combination, sizeof(unsigned int));
			}
		}

		//out->write((char*)(&combinationsUniq.size()), sizeof(int));
		int numDictionaries = 0;
		for (it = combinationsUniq.begin(); it != combinationsUniq.end();
				++it) {

			//create dictionary
			ModifiableDictionary* newDictionary = new PlainDictionary(spec);
			newDictionary->startProcessing();

			for (int n = 0; n < triples_positionsPerCombination.at(*it).size();
					n++) {
				int currentPosition =
						triples_positionsPerCombination.at(*it).at(n);
				TripleID tid = *itPositionTriple->getTripleFromPosition(
						currentPosition);

				//insert into triplesListsperCombination
				triplesListsperCombination.at(*it)->insert(tid);
				string subject = dictionary->idToString(tid.getSubject(),
						SUBJECT);
				string predicate = dictionary->idToString(tid.getPredicate(),
						PREDICATE);
				string object = dictionary->idToString(tid.getObject(), OBJECT);

				newDictionary->insert(subject, SUBJECT);
				newDictionary->insert(predicate, PREDICATE);
				newDictionary->insert(object, OBJECT);

				//dictionary->insert(triple.getPredicate(), PREDICATE);
				//dictionary->insert(triple.getObject(), OBJECT);

			}
			newDictionary->stopProcessing(NULL);
			Dictionary* dictionaryFourSections = new FourSectionDictionary(spec);
			dictionaryFourSections->import(newDictionary);
			delete newDictionary;

			ControlInformation controlInformation;
			controlInformation.clear();
			controlInformation.setType(DICTIONARY);

			if (singleFileOption) {
				dictionaryFourSections->save(*out, controlInformation, NULL);
			} else {
				ostream *outDict;
				ofstream outFDict;

				std::stringstream outfile;
				outfile << outputDirDictionaries << (*it) << ".dic";

				outFDict.open(outfile.str().c_str());
				outDict = &outFDict;
				dictionaryFourSections->save(*outDict, controlInformation,
				NULL);
			}
			numDictionaries++;

			// CREATE TRIPLES
			// --> replace the old IDs with the new IDs
			ModifiableTriples* oldtriples = triplesListsperCombination.at(*it);
			ModifiableTriples* newtriples = new TriplesList(spec);
			IteratorTripleID *it2 = oldtriples->searchAll();
			while (it2->hasNext()) {
				TripleID *triple = it2->next();
					//cout<<*triple<<endl;
				string subjString = dictionary->idToString(triple->getSubject(),
						SUBJECT);
				unsigned int newSubjID = dictionaryFourSections->stringToId(
						subjString, SUBJECT);
				triple->setSubject(newSubjID);
				

				string predString = dictionary->idToString(
						triple->getPredicate(), PREDICATE);
				unsigned int newPredID = dictionaryFourSections->stringToId(
						predString, PREDICATE);
				triple->setPredicate(newPredID);

				string objString = dictionary->idToString(triple->getObject(),
						OBJECT);
				unsigned int newObjID = dictionaryFourSections->stringToId(
						objString, OBJECT);
				triple->setObject(newObjID);
				newtriples->insert(*triple);

			}

			//sort and remove duplicates (just in case)
			oldtriples->sort(SPO, NULL);
			/*
			//test current triples to save:
			cout<<"test current triples to save:"<<endl;
			IteratorTripleID* itTid = newtriples->searchAll();
			while (itTid->hasNext()){
				cout<<*itTid->next()<<endl;
			}
			*/
			
			cout << "   Create Bitmap Triples and save Triples" << endl;
			BitmapTriples* bt = new BitmapTriples(spec);
			bt->load(*newtriples, NULL);
			ControlInformation ci;

			ci.clear();
			ci.setType(TRIPLES);

			if (singleFileOption) {
				bt->save(*out, ci, NULL);

			} else {
				ostream *outTriples;
				ofstream outFTriples;

				std::stringstream outfile;
				outfile << outputDirTriples << (*it) << ".bt";

				outFTriples.open(outfile.str().c_str());
				outTriples = &outFTriples;
				bt->save(*outTriples, ci, NULL);

			}

		}
		cout << "numDictionaries:" << numDictionaries << endl;
		cout << "* " << combinationsUniq.size() << " combinations *" << endl;





		

		*outTime << st << endl;

		if(outputTimeFile!="") {
			outFTime.close();
		}

		delete hdt; // Remember to delete instance when no longer needed!
	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

void ParseCallback::processTriple(hdt::TripleString& triple,
		unsigned long long pos) {
	//cout << "Triple String: " << triple << endl;
	count++;
	if ((count % 10000000) == 0)
		cout << count << " lines" << endl;

	string subject = triple.getSubject();
	string predicate = triple.getPredicate();
	string object = triple.getObject();
	//cout << "Result String: " << subject << ", " << predicate << ", " << object << endl;

	if (subject.at(0) == '<') {
		subject = subject.substr(1, subject.length() - 2);
	}
	if (object.at(0) == '<') {
		object = object.substr(1, object.length() - 2);
	}
	if (predicate.at(0) == '<') {
		predicate = predicate.substr(1, predicate.length() - 2);
	}
	//cout << "Result after filtering String: " << subject << ", " << predicate << ", " << object << endl;

	int subj = dictionary->stringToId(subject, SUBJECT);
	
	int pred = dictionary->stringToId(predicate, PREDICATE);

	
	int obj = dictionary->stringToId(object, OBJECT);

	//insert triple
	TripleID* tid = new TripleID(subj, pred, obj);
	//get position of the triple
	int positionTriple = itgetPositionTriple->getNumberofTriple(*tid);
	//mark this position in the graphk
	bitmapTriplesGraph->set(positionTriple, true);

	if (subj == 0 || pred == 0 || obj == 0) {
		cerr << "Error:" << subject + " " + predicate + " " + object << endl;
		exit(-1);
	}
	/*if ((count % 10000) == 0)
	 cout << subj << " " << pred << " " << obj << " ." << endl;
	 */

}

