/*
 * SPLIT IN VERSION C: NON OVERLAPPING DICTIONARIES AND OVERLAPPING TRIPLES
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <bitset>
#include <math.h>
#include "../src/libdcs/CSD_PFC.h"
#include "../src/libdcs/CSD.h"
#include "../src/util/StopWatch.hpp"

#include <HDTManager.hpp>
#include "../src/rdf/RDFParserNtriplesCallback.hpp"
#include "../src/bitsequence/BitSequence375.h"
#include "../src/triples/TriplesList.hpp"
#include "../src/triples/BitmapTriples.hpp"
#include <HDTEnums.hpp>

using namespace std;
using namespace hdt;

class ParseCallback: public RDFCallback {

private:
	long count;
	Dictionary *dictionary;
	BitSequence375* bitmapDictionariesGraph;
	ModifiableTriples* triplesListsGraph;
	BitSequence375* bitmapsSubjetsGraph;

public:
	ParseCallback(Dictionary *dictionary,
			BitSequence375* bitmapDictionariesGraph,
			BitSequence375* bitmapsSubjetsGraph,
			ModifiableTriples* triplesListsGraph) :
			dictionary(dictionary), count(0), bitmapDictionariesGraph(
					bitmapDictionariesGraph), bitmapsSubjetsGraph(
					bitmapsSubjetsGraph), triplesListsGraph(triplesListsGraph) {
	}
	;
	void processTriple(hdt::TripleString& triple, unsigned long long pos);
};

void help() {
	cout
			<< "$ splitHDT_C [options] <full_hdt_File> <number_of_partitions> <directory_With_Partitions>"
			<< endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-s\t\t\tSave current dictionary and exit" << endl;
	cout << "\t-o\t<outputSingleFile>\tSave split into one single file ."
			<< endl;
	cout
			<< "\t-d\t<outputDirDictionaries>\tSave dictionary split into several files in the outputDirDictionaries directory."
			<< endl;
	cout
			<< "\t-t\t<outputDirTriples>\tSave triples split into several files in the outputDirTriples directory."
			<< endl;
	cout << "\t-r\t<timeFile>\tSave the performance time to a file." << endl;
	//cout << "\t-v\tVerbose output" << endl;
}

class PositionIteratorUCharString: public IteratorUCharString {
private:
	std::vector<int> vector;
	Dictionary* dictionary;
	unsigned int pos;
public:
	PositionIteratorUCharString(std::vector<int> vector, Dictionary* dictionary) :
			pos(0), dictionary(dictionary), vector(vector) {
	}
	//PositionIteratorUCharString(std::vector<int> vector, Dictionary* dictionary) : vector(vector), dictionary(dictionary) pos(0) { }
	virtual ~PositionIteratorUCharString() {
	}

	virtual bool hasNext() {
		return pos < vector.size();
	}

	virtual unsigned char *next() {
		int id = vector[pos++];
		string ret = "";
		if (id <= dictionary->getNshared()) {
			ret = dictionary->idToString(id, SUBJECT);
		} else if (id <= dictionary->getNsubjects()) {
			ret = dictionary->idToString(id, SUBJECT);
		} else if (id
				> (dictionary->getNumberOfElements()
						- dictionary->getNpredicates())) {
			//	cout<<"id:"<<id<<endl;
			//	cout<<" dictionary->getNumberOfElements():"<< dictionary->getNumberOfElements()<<endl;
			//	cout << "id-(dictionary->getNumberOfElements()- dictionary->getNpredicates()):"
			//			<< (id-(dictionary->getNumberOfElements()
			//				- dictionary->getNpredicates()))<<endl;
			ret = dictionary->idToString(
					(id
							- (dictionary->getNumberOfElements()
									- dictionary->getNpredicates())),
					PREDICATE);
			//	cout<<"ret:"<<ret<<endl;
		} else {
			ret = dictionary->idToString(
					id - dictionary->getNsubjects() + dictionary->getNshared(),
					OBJECT);
		}
		//cerr<<ret<<endl;
		// fflush(stderr);
		/*if (ret.length()==0){
		 cerr<<"ret is 0 with id="<<id<<endl;
		 exit(-1);
		 }*/
		char* classVariableCharPointer = new char[strlen((char*) ret.c_str())
				+ 1];

		// strncpy((char*)(text+bytes), (char*)classVariableCharPointer, currentLength);
		strcpy(classVariableCharPointer, (const char*) ret.c_str());
		return (unsigned char*) classVariableCharPointer;
		//return (unsigned char *) ret.c_str();
	}
};

int main(int argc, char *argv[]) {

	StopWatch st;

	string query, inputFile, outputsingleFile, outputDirDictionaries,
			outputDirTriples;
	string outputTimeFile;
	bool savecurrentDictionary = false;
	int c;

	while ((c = getopt(argc, argv, "ho:d:t:sr:")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'o':
			outputsingleFile = optarg;
			break;
		case 's':
			savecurrentDictionary = true;
			break;
		case 'd':
			outputDirDictionaries = optarg;
			break;
		case 't':
			outputDirTriples = optarg;
			break;
		case 'r':
			outputTimeFile = optarg;
			break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	if (argc - optind < 1) {
		cout << "ERROR: You must supply an HDT File" << endl << endl;
		help();
		return 1;
	}

	inputFile = argv[optind];

	if (argc < 4) {
		cerr << "Incorrect number of parameters!" << endl;
		help();
		exit(0);
	}


	try {

		//prepare the time output
		ostream *outTime;
		ofstream outFTime;

		if(outputTimeFile!="") {
			outFTime.open(outputTimeFile.c_str());
			outTime = &outFTime;
		} else {
			outTime = &cout;
		}

		// Load HDT main file
		HDT *hdt = HDTManager::mapHDT(inputFile.c_str());
		cout << "Opening " << inputFile << endl;
		Dictionary *dictionary = hdt->getDictionary();

		//save dictionary just to compare the size

		if (savecurrentDictionary) {
			std::ostream* outDict;
			ofstream outD;
			outD.open("OutDictionary.dic");
			outDict = &outD;
			dictionary->save(*outDict, *new ControlInformation(), NULL);
			exit(0);
		}
		int numVersions = atoi(argv[optind + 1]);

		vector<BitSequence375*> bitmapsDictionariesGraph(numVersions);
		vector<BitSequence375*> bitmapsSubjetsGraph(numVersions); //just marking the real subjects
		vector<ModifiableTriples*> triplesListsGraph(numVersions);
		// Process
		string configFile = "";
		string options = "";
		HDTSpecification spec(configFile);

		spec.setOptions(options);
		for (int i = 0; i < numVersions; i++) {
			bitmapsDictionariesGraph.at(i) = new BitSequence375();
			bitmapsSubjetsGraph.at(i) = new BitSequence375();
			triplesListsGraph.at(i) = new TriplesList(spec);
		}

		cout << "Total elements:" << dictionary->getNumberOfElements() << endl;

		RDFParserNtriplesCallback* parser = new RDFParserNtriplesCallback();

		RDFNotation notation = NTRIPLES;

		string dirPartitions = argv[optind + 2];

		/*
		 * FIRST PASS, ADD THE 1'S TO THE BITMAPS
		 */

		for (int i = 1; i < numVersions + 1; i++) {

			ParseCallback *callback = new ParseCallback(dictionary,
					bitmapsDictionariesGraph.at(i - 1),
					bitmapsSubjetsGraph.at(i - 1), triplesListsGraph.at(i - 1));
			stringstream ss;
			ss << i;
			string filename = dirPartitions + ss.str() + ".nt";
			cout << filename << endl;
			parser->doParse((const char*) filename.c_str(), "", notation,
					callback);
		}
		delete parser; // Remember to delete iterator to avoid memory leaks!

		/*
		 * COMPUTE THE COMBINATIONS
		 */
		vector<unsigned int> combinations;
		set<unsigned int> combinationsUniq;
		vector<vector<int> > so_positionsPerCombination(pow(2, numVersions));
		vector<vector<int> > s_positionsPerCombination(pow(2, numVersions));
		vector<vector<int> > o_positionsPerCombination(pow(2, numVersions));
		vector<vector<int> > p_positionsPerCombination(pow(2, numVersions));
		for (int i = 0; i < dictionary->getNumberOfElements(); i++) {
			string s = "";
			for (int j = 0; j < numVersions; j++) {
				s = (bitmapsDictionariesGraph.at(j)->access(i) ? "1" : "0") + s;
			}
			int result = std::bitset<32>(s).to_ulong();
			/*if (result==0){
			 cout<<"found 0 in position:"<<i<<endl;
			 exit(-1);
			 }*/
			//	cout<<"s is:"<<s<<endl;
			//	cout<<"result is:"<<result<<endl;
			combinations.push_back(result);
			combinationsUniq.insert(result);
			//store the position in the corresponding vector
			if ((i + 1) <= dictionary->getNshared()) {
				so_positionsPerCombination.at(result).push_back(i + 1); //+1 because there is not the id =0
			} else if ((i + 1) <= dictionary->getNsubjects()) {
				s_positionsPerCombination.at(result).push_back(i + 1); //+1 because there is not the id =0
			} else if ((i + 1)
					> (dictionary->getNumberOfElements()
							- dictionary->getNpredicates())) {
				p_positionsPerCombination.at(result).push_back(i + 1); //+1 because there is not the id =0
			} else {
				o_positionsPerCombination.at(result).push_back(i + 1); //+1 because there is not the id =0
			}
		}
		std::set<unsigned int>::iterator it;
		cout << "* " << combinationsUniq.size() << " combinations *" << endl;
		ostream *out;
		ofstream outF;

		bool singleFileOption = false;
		if (outputsingleFile != "") {
			singleFileOption = true;
			outF.open(outputsingleFile.c_str());
			out = &outF;
		} else {
			out = &cout;
			if (outputDirDictionaries == "" || outputDirTriples == "") {
				cerr
						<< "Please provide one output method (-o for single mode output or (-d and -t) for multiple files)"
						<< endl;
				help();
				exit(-1);
			}
		}
		//singleFileOption:concatenate all dictionaries&triples in a file

		// CREATE SUBDICTIONARIES (FOR EACH COMBINATION WE WILL HAVE 1 SUBDICTIONARY SUBJECT-OBJECT, 1 UNIQ SUBJECT, 1 UNIQ OBJECT AND 1 PREDICATE

		//metadata: save the number of versions and the number of combinations

		//	*out<<(unsigned int)numVersions;
		//	*out<<(unsigned int)combinationsUniq.size();

		if (singleFileOption) {
			out->write((char*) &numVersions, sizeof(unsigned int));
			unsigned int numCombinations = combinationsUniq.size();
			out->write((char*) &numCombinations, sizeof(unsigned int));
			// write combinations
			for (it = combinationsUniq.begin(); it != combinationsUniq.end();
					++it) {
				unsigned int combination = *it;
				out->write((char*) &combination, sizeof(unsigned int));
			}
		}

		//out->write((char*)(&combinationsUniq.size()), sizeof(int));
		int numDictionaries = 0;
		for (it = combinationsUniq.begin(); it != combinationsUniq.end();
				++it) {
			//	cerr <<"Combination:"<< *it << endl;
			//if (*it>10)
			//	exit(0);
			//create one PFC dictionary per combination
			// SO
			//		cerr<<"SO:"<<endl;
			PositionIteratorUCharString* iteratePositions =
					new PositionIteratorUCharString(
							so_positionsPerCombination.at(*it), dictionary);

			csd::CSD* dict_so = new csd::CSD_PFC(iteratePositions, 16, NULL);
			//	cerr<<"End SO"<<endl;

			//test SO
			/*	for (int k = 0; k < dict_so->getLength(); k++) {
			 cerr << "k so:" << k << endl;
			 cerr << "dict_so->extract(k + 1):" << dict_so->extract(k + 1)
			 << "." << endl;

			 }*/

			numDictionaries++;

			//S
			//cout<<"S:"<<endl;
			iteratePositions = new PositionIteratorUCharString(
					s_positionsPerCombination.at(*it), dictionary);

			csd::CSD* dict_s = new csd::CSD_PFC(iteratePositions, 16, NULL);
			//cout<<"End S"<<endl;

			numDictionaries++;
			//O
			//cout<<"O:"<<endl;
			iteratePositions = new PositionIteratorUCharString(
					o_positionsPerCombination.at(*it), dictionary);

			csd::CSD* dict_o = new csd::CSD_PFC(iteratePositions, 16, NULL);
			//cout<<"End O"<<endl;

			numDictionaries++;
			//P
			//cout<<"P:"<<endl;
			iteratePositions = new PositionIteratorUCharString(
					p_positionsPerCombination.at(*it), dictionary);

			csd::CSD* dict_p = new csd::CSD_PFC(iteratePositions, 16, NULL);
			//cout<<"End P"<<endl;

			if (singleFileOption) {
				dict_so->save(*out);
				dict_s->save(*out);
				dict_o->save(*out);
				dict_p->save(*out);
			} else {
				ostream *outDict;
				ofstream outFDict;

				std::stringstream outfile;
				outfile << outputDirDictionaries << (*it) << ".dic";

				outFDict.open(outfile.str().c_str());
				outDict = &outFDict;
				dict_so->save(*outDict);
				dict_s->save(*outDict);
				dict_o->save(*outDict);
				dict_p->save(*outDict);
			}
			numDictionaries++;
		}
		cout << "numDictionaries:" << numDictionaries << endl;
		cout << "* " << combinationsUniq.size() << " combinations *" << endl;

		// CREATE TRIPLES
		// --> replace the old IDs with the new IDs

		cout << "First, keep number of SO, S and O of each version" << endl;
		// first, keep number of SO, S and O of each version
		vector<int> numSOperGraph;
		vector<int> numSperGraph;
		vector<int> numSO_S_OperGraph;
		vector<int> numPperGraph;
		for (int i = 0; i < numVersions; i++) {
			//	cout << "i:" << i << endl;
			numSOperGraph.push_back(
					bitmapsDictionariesGraph.at(i)->rank1(
							dictionary->getNshared() - 1));
			

			numSperGraph.push_back(
					bitmapsDictionariesGraph.at(i)->rank1(
							dictionary->getNsubjects() - 1));
			

			numSO_S_OperGraph.push_back(
					bitmapsDictionariesGraph.at(i)->rank1(
							dictionary->getNumberOfElements()
									- dictionary->getNpredicates() - 1));
			

		}

		cout << "Iterate triples per version" << endl;

		for (int i = 0; i < numVersions; i++) {
			cout << "  Version:" << i << endl;
			ModifiableTriples* oldtriples = triplesListsGraph.at(i);
			ModifiableTriples* newtriples = new TriplesList(spec);
			BitSequence375* currectBitmapDictionaryGraph =
					bitmapsDictionariesGraph.at(i);
			BitSequence375* currectBitmapSubjectGraph = bitmapsSubjetsGraph.at(
					i);
			
			IteratorTripleID *it = oldtriples->searchAll();
			while (it->hasNext()) {
				TripleID *triple = it->next();
				
				//use the ID of the present subjects (not the SO because one SO can be there in a partition just because of an object)
				triple->setSubject(
						currectBitmapSubjectGraph->rank1(
								triple->getSubject() - 1));

				//new PredicateID is the number of 1s in  current Bitmap up to that position minus the number of previous SO, S and O in the graph
				triple->setPredicate(
						currectBitmapDictionaryGraph->rank1(
								triple->getPredicate()
										+ (dictionary->getNumberOfElements()
												- dictionary->getNpredicates())
										- 1) - numSO_S_OperGraph.at(i));
				if (triple->getObject() <= dictionary->getNshared()) {
					//new ObjectID is the number of 1s in  current Bitmap up to that position
					triple->setObject(
							currectBitmapDictionaryGraph->rank1(
									triple->getObject() - 1));

				} else {
					//cout << "object:" << triple->getObject() << endl;
					//new ObjectID is the number of 1s in  current Bitmap up to that object plus the number of unique subjects, that is the number of total subject minus the number of shared SO
					triple->setObject(
							currectBitmapDictionaryGraph->rank1(
									triple->getObject()
											+ dictionary->getNsubjects()
											- dictionary->getNshared() - 1)
									- numSperGraph.at(i) + numSOperGraph.at(i));
					
				}
				
				//insert new triple
				newtriples->insert(*triple);

				
			}
			//sort and remove duplicates (just in case)
			newtriples->sort(SPO, NULL);
			newtriples->removeDuplicates(NULL); 
			cout << "   Create Bitmap Triples and save Triples" << endl;
			BitmapTriples* bt = new BitmapTriples(spec);
			bt->load(*newtriples, NULL);
			ControlInformation ci;

			ci.clear();
			ci.setType(TRIPLES);

			//moved after computing bitmaps of SO to put all together in the output
			//bt->save(*out, ci, NULL);

			
			// BUILD FINAL BITMAP SUBJECTS WITHOUT GAPS
			// Mark which of the new SO is a real Subject
			BitSequence375* actualSubjectsofSO = new BitSequence375();
			int numgraphSubjects = 0;
			for (int k = 0; k < dictionary->getNshared(); k++) {
				if (currectBitmapDictionaryGraph->access(k) == true) {
					actualSubjectsofSO->set(numgraphSubjects,
							currectBitmapSubjectGraph->access(k));
					numgraphSubjects++;
				}
			}
			//cout<<"numgraphSubjects:"<<numgraphSubjects<<endl;

			if (singleFileOption) {
				bt->save(*out, ci, NULL);
				actualSubjectsofSO->save(*out);
			} else {
				ostream *outTriples;
				ofstream outFTriples;

				std::stringstream outfile;
				outfile << outputDirTriples << (i + 1) << ".bt";

				outFTriples.open(outfile.str().c_str());
				outTriples = &outFTriples;
				bt->save(*outTriples, ci, NULL);
				actualSubjectsofSO->save(*outTriples);

			}

		}

		

		*outTime << st << endl;

		if(outputTimeFile!="") {
			outFTime.close();
		}

		delete hdt; // Remember to delete instance when no longer needed!
	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

void ParseCallback::processTriple(hdt::TripleString& triple,
		unsigned long long pos) {
	//cout << "Triple String: " << triple << endl;
	count++;
	if ((count % 10000000) == 0)
		cout << count << " lines" << endl;

	string subject = triple.getSubject();
	string predicate = triple.getPredicate();
	string object = triple.getObject();
	//cout << "Result String: " << subject << ", " << predicate << ", " << object << endl;

	if (subject.at(0) == '<') {
		subject = subject.substr(1, subject.length() - 2);
	}
	if (object.at(0) == '<') {
		object = object.substr(1, object.length() - 2);
	}
	if (predicate.at(0) == '<') {
		predicate = predicate.substr(1, predicate.length() - 2);
	}
	//cout << "Result after filtering String: " << subject << ", " << predicate << ", " << object << endl;

	int subj = dictionary->stringToId(subject, SUBJECT);

	if (subj == 10) {
		cout << "subject is TEN!" << endl;
	}
	/*
	 * PUT SUBJECT IN THE BEGINING
	 */

	bitmapDictionariesGraph->set(subj - 1, true);

	//mark subject as present
	bitmapsSubjetsGraph->set(subj - 1, true);

	
	int pred = dictionary->stringToId(predicate, PREDICATE);
	/*
	 * PUT at the end
	 */
	bitmapDictionariesGraph->set(
			(dictionary->getNumberOfElements() - dictionary->getNpredicates())
					+ pred - 1, true);

	
	int obj = dictionary->stringToId(object, OBJECT);
	
	/*
	 * PUT NON SHARED OBJECTS AFTER THE SUBJECTS
	 */
	if (obj <= dictionary->getNshared()) {
		bitmapDictionariesGraph->set(obj - 1, true);
	} else {
		bitmapDictionariesGraph->set(
				dictionary->getNsubjects() + obj - dictionary->getNshared() - 1,
				true);
	}

	//insert triple
	TripleID* tid = new TripleID(subj, pred, obj);
	
	triplesListsGraph->insert(*tid);

	if (subj == 0 || pred == 0 || obj == 0) {
		cerr << "Error:" << subject + " " + predicate + " " + object << endl;
		exit(-1);
	}
	/*if ((count % 10000) == 0)
	 cout << subj << " " << pred << " " << obj << " ." << endl;
	 */

}

