/*
 * SPLIT IN VERSION D: NON OVERLAPPING DICTIONARIES AND NON OVERLAPPING TRIPLES
 */

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <string>
#include <bitset>
#include <math.h>
#include "../src/libdcs/CSD_PFC.h"
#include "../src/libdcs/CSD.h"
#include "../src/util/StopWatch.hpp"

#include <HDTManager.hpp>
#include "../src/rdf/RDFParserNtriplesCallback.hpp"
#include "../src/bitsequence/BitSequence375.h"
#include "../src/triples/TriplesList.hpp"
#include "../src/triples/BitmapTriples.hpp"
#include "../src/triples/CompactTriples.hpp"
#include "../src/triples/PlainTriples.hpp"
#include <HDTEnums.hpp>

#include "toolsSplit.hpp"

using namespace std;
using namespace hdt;

class ParseCallback: public RDFCallback {

private:
	long count;
	Dictionary *dictionary;
	BitSequence375* bitmapDictionariesGraph;
	BitSequence375* bitmapTriplesGraph;
	BitSequence375* bitmapsSubjetsGraph;
	IteratorGetPosition* itgetPositionTriple;

public:
	ParseCallback(Dictionary *dictionary,
			BitSequence375* bitmapDictionariesGraph,
			BitSequence375* bitmapsSubjetsGraph,
			BitSequence375* bitmapTriplesGraph,
			IteratorGetPosition* itgetPositionTriple) :
			dictionary(dictionary), count(0), bitmapDictionariesGraph(
					bitmapDictionariesGraph), bitmapsSubjetsGraph(
					bitmapsSubjetsGraph), bitmapTriplesGraph(
					bitmapTriplesGraph), itgetPositionTriple(
					itgetPositionTriple) {
	}
	;
	void processTriple(hdt::TripleString& triple, unsigned long long pos);
};

void help() {
	cout
			<< "$ splitHDT_D [options] <full_hdt_File> <number_of_partitions> <directory_With_Partitions>"
			<< endl;
	cout << "\t-h\t\t\tThis help" << endl;
	cout << "\t-s\t\t\tSave current dictionary and exit" << endl;
	cout << "\t-o\t<outputSingleFile>\tSave split into one single file ."
			<< endl;
	cout
			<< "\t-d\t<outputDirDictionaries>\tSave dictionary split into several files in the outputDirDictionaries directory."
			<< endl;
	cout
			<< "\t-t\t<outputDirTriples>\tSave triples split into several files in the outputDirTriples directory."
			<< endl;

	cout << "\t-r\t<timeFile>\tSave the performance time to a file."
					<< endl;
	//cout << "\t-v\tVerbose output" << endl;
}

class PositionIteratorUCharString: public IteratorUCharString {
private:
	std::vector<int> vector;
	Dictionary* dictionary;
	unsigned int pos;
public:
	PositionIteratorUCharString(std::vector<int> vector, Dictionary* dictionary) :
			pos(0), dictionary(dictionary), vector(vector) {
	}
	//PositionIteratorUCharString(std::vector<int> vector, Dictionary* dictionary) : vector(vector), dictionary(dictionary) pos(0) { }
	virtual ~PositionIteratorUCharString() {
	}

	virtual bool hasNext() {
		return pos < vector.size();
	}

	virtual unsigned char *next() {
		int id = vector[pos++];
		string ret = "";
		if (id <= dictionary->getNshared()) {
			ret = dictionary->idToString(id, SUBJECT);
		} else if (id <= dictionary->getNsubjects()) {
			ret = dictionary->idToString(id, SUBJECT);
		} else if (id
				> (dictionary->getNumberOfElements()
						- dictionary->getNpredicates())) {
			//	cout<<"id:"<<id<<endl;
			//	cout<<" dictionary->getNumberOfElements():"<< dictionary->getNumberOfElements()<<endl;
			//	cout << "id-(dictionary->getNumberOfElements()- dictionary->getNpredicates()):"
			//			<< (id-(dictionary->getNumberOfElements()
			//				- dictionary->getNpredicates()))<<endl;
			ret = dictionary->idToString(
					(id
							- (dictionary->getNumberOfElements()
									- dictionary->getNpredicates())),
					PREDICATE);
			//	cout<<"ret:"<<ret<<endl;
		} else {
			ret = dictionary->idToString(
					id - dictionary->getNsubjects() + dictionary->getNshared(),
					OBJECT);
		}
		//cerr<<ret<<endl;
		// fflush(stderr);
		/*if (ret.length()==0){
		 cerr<<"ret is 0 with id="<<id<<endl;
		 exit(-1);
		 }*/
		char* classVariableCharPointer = new char[strlen((char*) ret.c_str())
				+ 1];

		// strncpy((char*)(text+bytes), (char*)classVariableCharPointer, currentLength);
		strcpy(classVariableCharPointer, (const char*) ret.c_str());
		return (unsigned char*) classVariableCharPointer;
		//return (unsigned char *) ret.c_str();
	}
};

IteratorGetPosition::IteratorGetPosition(BitmapTriples *trip) :
		triples(trip), adjY(trip->arrayY, trip->bitmapY), adjZ(trip->arrayZ,
				trip->bitmapZ) {

	minY = minZ = 0;
	patX = patY = patZ = 0;
	x = y = z = maxY = maxZ = posZ = posY = 0;

}
int IteratorGetPosition::getNumberofTriple(TripleID &pat) {

	unsigned int retID = 0;
	// assume the order of trip and pat is SPO
	unsigned int patX = pat.getSubject();
	unsigned int patY = pat.getPredicate();
	unsigned int patZ = pat.getObject();

	if (patX != 0 && patY != 0 && patZ != 0) {
		// S P O
		try {
			minY = adjY.find(patX - 1, patY);
			maxY = minY + 1;

			minZ = adjZ.find(minY, patZ);
			retID = minZ;
			maxZ = minZ + 1;

		} catch (const char *ex) {
			// Item not found in list, no results.
			minY = minZ = maxY = maxZ = 0;
		}

		x = patX;
	}
	return retID;
}

TripleID* IteratorGetPosition::getTripleFromPosition(unsigned int position) {

	posY = adjZ.findListIndex(position);

	z = adjZ.get(position);
	y = adjY.get(posY);
	x = adjY.findListIndex(posY) + 1;

	return new TripleID(x, y, z);

}

int main(int argc, char *argv[]) {

	StopWatch st;

	string query, inputFile, outputsingleFile, outputDirDictionaries,
			outputDirTriples;
	string outputTimeFile;
	bool savecurrentDictionary = false;
	int c;

	while ((c = getopt(argc, argv, "ho:d:t:sr:")) != -1) {
		switch (c) {
		case 'h':
			help();
			break;
		case 'o':
			outputsingleFile = optarg;
			break;
		case 's':
			savecurrentDictionary = true;
			break;
		case 'd':
			outputDirDictionaries = optarg;
			break;
		case 't':
			outputDirTriples = optarg;
			break;
		case 'r':
					outputTimeFile = optarg;
					break;
		default:
			cout << "ERROR: Unknown option" << endl;
			help();
			return 1;
		}
	}

	if (argc - optind < 1) {
		cout << "ERROR: You must supply an HDT File" << endl << endl;
		help();
		return 1;
	}

	inputFile = argv[optind];

	if (argc < 4) {
		cerr << "Incorrect number of parameters!" << endl;
		help();
		exit(0);
	}

	try {

		//prepare the time output
		ostream *outTime;
		ofstream outFTime;

		if(outputTimeFile!="") {
			outFTime.open(outputTimeFile.c_str());
			outTime = &outFTime;
		} else {
			outTime = &cout;
		}

		// Load HDT main file
		HDT *hdt = HDTManager::mapHDT(inputFile.c_str());
		cout << "Opening " << inputFile << endl;
		Dictionary *dictionary = hdt->getDictionary();
		Triples *triples = hdt->getTriples();
		// iterator for the position
		IteratorGetPosition* itPositionTriple = new IteratorGetPosition(
				(BitmapTriples*) (hdt->getTriples()));

		//save dictionary just to compare the size

		if (savecurrentDictionary) {
			std::ostream* outDict;
			ofstream outD;
			outD.open("OutDictionary.dic");
			outDict = &outD;
			dictionary->save(*outDict, *new ControlInformation(), NULL);
			exit(0);
		}
		int numVersions = atoi(argv[optind + 1]);

		vector<BitSequence375*> bitmapsDictionariesGraph(numVersions);
		vector<BitSequence375*> bitmapsSubjetsGraph(numVersions); //just marking the real subjects
		vector<BitSequence375*> bitmapsTriplesGraph(numVersions);
		// Process
		string configFile = "";
		string options = "";
		HDTSpecification spec(configFile);

		spec.setOptions(options);
		for (int i = 0; i < numVersions; i++) {
			bitmapsDictionariesGraph.at(i) = new BitSequence375();
			bitmapsSubjetsGraph.at(i) = new BitSequence375();

			bitmapsTriplesGraph.at(i) = new BitSequence375();
		}

		cout << "Total elements:" << dictionary->getNumberOfElements() << endl;

		RDFParserNtriplesCallback* parser = new RDFParserNtriplesCallback();

		RDFNotation notation = NTRIPLES;

		string dirPartitions = argv[optind + 2];

		/*
		 * FIRST PASS, ADD THE 1'S TO THE BITMAPS
		 */

		for (int i = 1; i < numVersions + 1; i++) {

			ParseCallback *callback = new ParseCallback(dictionary,
					bitmapsDictionariesGraph.at(i - 1),
					bitmapsSubjetsGraph.at(i - 1),
					bitmapsTriplesGraph.at(i - 1), itPositionTriple);
			stringstream ss;
			ss << i;
			string filename = dirPartitions + ss.str() + ".nt";
			cout << filename << endl;
			parser->doParse((const char*) filename.c_str(), "", notation,
					callback);
		}
		delete parser; // Remember to delete iterator to avoid memory leaks!

		/*
		 * COMPUTE THE COMBINATIONS
		 */
		vector<unsigned int> combinationsDictionaries;
		set<unsigned int> combinationsUniqDictionaries;
		vector<vector<int> > so_positionsPerCombination(pow(2, numVersions));
		vector<vector<int> > s_positionsPerCombination(pow(2, numVersions));
		vector<vector<int> > o_positionsPerCombination(pow(2, numVersions));
		vector<vector<int> > p_positionsPerCombination(pow(2, numVersions));
		vector<vector<int> > triples_positionsPerCombination(
				pow(2, numVersions));
		vector<ModifiableTriples*> triplesListsperCombination(
				pow(2, numVersions));

		//initialize triplesListsperCombination;
		for (int i = 0; i < pow(2, numVersions); i++) {
			triplesListsperCombination.at(i) = new TriplesList(spec);
		}

		for (int i = 0; i < dictionary->getNumberOfElements(); i++) {
			string s = "";
			for (int j = 0; j < numVersions; j++) {
				s = (bitmapsDictionariesGraph.at(j)->access(i) ? "1" : "0") + s;
			}
			int result = std::bitset<32>(s).to_ulong();
			/*if (result==0){
			 cout<<"found 0 in position:"<<i<<endl;
			 exit(-1);
			 }*/
			//	cout<<"s is:"<<s<<endl;
			//	cout<<"result is:"<<result<<endl;
			combinationsDictionaries.push_back(result);
			combinationsUniqDictionaries.insert(result);
			//store the position in the corresponding vector
			if ((i + 1) <= dictionary->getNshared()) {
				so_positionsPerCombination.at(result).push_back(i + 1); //+1 because there is not the id =0
			} else if ((i + 1) <= dictionary->getNsubjects()) {
				s_positionsPerCombination.at(result).push_back(i + 1); //+1 because there is not the id =0
			} else if ((i + 1)
					> (dictionary->getNumberOfElements()
							- dictionary->getNpredicates())) {
				p_positionsPerCombination.at(result).push_back(i + 1); //+1 because there is not the id =0
			} else {
				o_positionsPerCombination.at(result).push_back(i + 1); //+1 because there is not the id =0
			}

		}
		vector<unsigned int> combinationsTriples;
		set<unsigned int> combinationsUniqTriples;
		//mark triples
		for (int i = 0; i < triples->getNumberOfElements(); i++) {
			string s = "";
			for (int j = 0; j < numVersions; j++) {
				s = (bitmapsTriplesGraph.at(j)->access(i) ? "1" : "0") + s;
			}
			int result = std::bitset<32>(s).to_ulong();
			/*if (result==0){
			 cout<<"found 0 in position:"<<i<<endl;
			 exit(-1);
			 }*/
			//	cout<<"s is:"<<s<<endl;
			//	cout<<"result is:"<<result<<endl;
			combinationsTriples.push_back(result);
			combinationsUniqTriples.insert(result);
			//store the position in the corresponding vector

			triples_positionsPerCombination.at(result).push_back(i);
		}

		std::set<unsigned int>::iterator it;
		cout << "* " << combinationsUniqDictionaries.size()
				<< " combinations Dictionary*" << endl;
		cout << "* " << combinationsUniqTriples.size()
				<< " combinations Triples*" << endl;
		ostream *out;
		ofstream outF;

		bool singleFileOption = false;
		if (outputsingleFile != "") {
			singleFileOption = true;
			outF.open(outputsingleFile.c_str());
			out = &outF;
		} else {
			out = &cout;
			if (outputDirDictionaries == "" || outputDirTriples == "") {
				cerr
						<< "Please provide one output method (-o for single mode output or (-d and -t) for multiple files)"
						<< endl;
				help();
				exit(-1);
			}
		}
		//singleFileOption:concatenate all dictionaries&triples in a file

		// CREATE SUBDICTIONARIES (FOR EACH COMBINATION WE WILL HAVE 1 SUBDICTIONARY SUBJECT-OBJECT, 1 UNIQ SUBJECT, 1 UNIQ OBJECT AND 1 PREDICATE

		//metadata: save the number of versions and the number of combinations

		//	*out<<(unsigned int)numVersions;
		//	*out<<(unsigned int)combinationsUniq.size();

		if (singleFileOption) {
			out->write((char*) &numVersions, sizeof(unsigned int));
			unsigned int numCombinationsDictionaries =
					combinationsUniqDictionaries.size();
			out->write((char*) &numCombinationsDictionaries,
					sizeof(unsigned int));
			unsigned int numCombinationsTriples =
					combinationsUniqTriples.size();
			out->write((char*) &numCombinationsTriples, sizeof(unsigned int));
			//I HAVE TO READ THIS IN THE READ_D BECAUSE IN READ_C I was not writing the combination of triples  too

			// write combinations
			for (it = combinationsUniqDictionaries.begin();
					it != combinationsUniqDictionaries.end(); ++it) {
				unsigned int combination = *it;
				out->write((char*) &combination, sizeof(unsigned int));
			}
			// write combinations
			for (it = combinationsUniqTriples.begin();
					it != combinationsUniqTriples.end(); ++it) {
				unsigned int combination = *it;
				out->write((char*) &combination, sizeof(unsigned int));
			}
		}

		//out->write((char*)(&combinationsUniq.size()), sizeof(int));
		int numDictionaries = 0;

		// I need to keep the dictionaries of the combinations
		vector<csd::CSD*> so_dicts(pow(2, numVersions));
		vector<csd::CSD*> s_dicts(pow(2, numVersions));
		vector<csd::CSD*> o_dicts(pow(2, numVersions));
		vector<csd::CSD*> p_dicts(pow(2, numVersions));

		for (it = combinationsUniqDictionaries.begin();
				it != combinationsUniqDictionaries.end(); ++it) {
			//	cerr <<"Combination:"<< *it << endl;
			//if (*it>10)
			//	exit(0);
			//create one PFC dictionary per combination
			// SO
			//		cerr<<"SO:"<<endl;
			PositionIteratorUCharString* iteratePositions =
					new PositionIteratorUCharString(
							so_positionsPerCombination.at(*it), dictionary);

			csd::CSD* dict_so = new csd::CSD_PFC(iteratePositions, 16, NULL);
			//	cerr<<"End SO"<<endl;

			//test SO
			/*	for (int k = 0; k < dict_so->getLength(); k++) {
			 cerr << "k so:" << k << endl;
			 cerr << "dict_so->extract(k + 1):" << dict_so->extract(k + 1)
			 << "." << endl;

			 }*/

			numDictionaries++;

			//S
			//cout<<"S:"<<endl;
			iteratePositions = new PositionIteratorUCharString(
					s_positionsPerCombination.at(*it), dictionary);

			csd::CSD* dict_s = new csd::CSD_PFC(iteratePositions, 16, NULL);
			//cout<<"End S"<<endl;

			numDictionaries++;
			//O
			//cout<<"O:"<<endl;
			iteratePositions = new PositionIteratorUCharString(
					o_positionsPerCombination.at(*it), dictionary);

			csd::CSD* dict_o = new csd::CSD_PFC(iteratePositions, 16, NULL);
			//cout<<"End O"<<endl;

			numDictionaries++;
			//P
			//cout<<"P:"<<endl;
			iteratePositions = new PositionIteratorUCharString(
					p_positionsPerCombination.at(*it), dictionary);

			csd::CSD* dict_p = new csd::CSD_PFC(iteratePositions, 16, NULL);
			//cout<<"End P"<<endl;

			if (singleFileOption) {
				dict_so->save(*out);
				dict_s->save(*out);
				dict_o->save(*out);
				dict_p->save(*out);
			} else {
				ostream *outDict;
				ofstream outFDict;

				std::stringstream outfile;
				outfile << outputDirDictionaries << (*it) << ".dic";

				outFDict.open(outfile.str().c_str());
				outDict = &outFDict;
				dict_so->save(*outDict);
				dict_s->save(*outDict);
				dict_o->save(*outDict);
				dict_p->save(*outDict);
			}

			//save dictionaries
			so_dicts.at(*it) = dict_so;
			s_dicts.at(*it) = dict_s;
			o_dicts.at(*it) = dict_o;
			p_dicts.at(*it) = dict_p;
			numDictionaries++;
		}
		cout << "numDictionaries:" << numDictionaries << endl;
		cout << "* " << combinationsUniqDictionaries.size()
				<< " combinations Dictionary*" << endl;
		cout << "* " << combinationsUniqTriples.size()
				<< " combinations Triples*" << endl;

		// CREATE TRIPLES
		// --> replace the old IDs with the new IDs

		/*******
		 * CHANGEEEEEEEEEEEEEEEEEEEEEEEEEEE
		 */
		cerr << "THE FOLLOWING HAS CHANGED TO CONSIDER THE HIERARCHIE" << endl;

		//cout << "First, keep number of SO, S and O of each version" << endl;
		// first, keep number of SO, S and O of each version
		vector<int> numSOperGraph;
		vector<int> numSperGraph;
		vector<int> numSO_S_OperGraph;
		vector<int> numPperGraph;
		for (int i = 0; i < numVersions; i++) {
			//	cout << "i:" << i << endl;
			numSOperGraph.push_back(
					bitmapsDictionariesGraph.at(i)->rank1(
							dictionary->getNshared() - 1));
			//cout << "numSOperGraph:"
			//		<< bitmapsDictionariesGraph.at(i)->rank1(
			//				dictionary->getNshared() - 1) << endl;

			numSperGraph.push_back(
					bitmapsDictionariesGraph.at(i)->rank1(
							dictionary->getNsubjects() - 1));
			//	cout << "numSperGraph:"
			//			<< bitmapsDictionariesGraph.at(i)->rank1(
			//					dictionary->getNsubjects() - 1) << endl;

			numSO_S_OperGraph.push_back(
					bitmapsDictionariesGraph.at(i)->rank1(
							dictionary->getNumberOfElements()
									- dictionary->getNpredicates() - 1));
			//	cout << "numSO_S_OperGraph:"
			//			<< bitmapsDictionariesGraph.at(i)->rank1(
			//					dictionary->getNumberOfElements()
			//							- dictionary->getNpredicates() - 1) << endl;

		}

		/*************************
		 * END NEED OF CHANGE
		 **********************/

		/*
		 * BEGIN DICTIONARY TEST
		 */

	/*	cout << "TEST PRINT ORIGINAL DICTIONARY" << endl;

		cout << "------------" << endl;
		cout << "SO:" << endl;
		IteratorUCharString* itElements = dictionary->getShared();
		int id = 1;
		while (itElements->hasNext()) {
			cout << id << ": " << itElements->next() << endl;
			id++;
		}
		cout << endl << endl;

		cout << "S:" << endl;
		itElements = dictionary->getSubjects();
		id = 1;
		while (itElements->hasNext()) {
			cout << id << ": " << itElements->next() << endl;
			id++;
		}
		cout << endl << endl;

		cout << "O:" << endl;
		itElements = dictionary->getObjects();
		id = 1;
		while (itElements->hasNext()) {
			cout << id << ": " << itElements->next() << endl;
			id++;
		}
		cout << endl << endl;

		cout << "P:" << endl;
		itElements = dictionary->getPredicates();
		id = 1;
		while (itElements->hasNext()) {
			cout << id << ": " << itElements->next() << endl;
			id++;
		}
		cout << endl << endl;

		cout << "TEST PRINT DICTIONARIES COMBINATIONS" << endl;
		//test print dictionaries
		for (it = combinationsUniqDictionaries.begin();
				it != combinationsUniqDictionaries.end(); ++it) {
			cout << " COMBINATION:" << *it << endl;
			cout << "------------" << endl;
			cout << "SO:" << endl;
			IteratorUCharString* itElements = so_dicts.at(*it)->listAll();
			int id = 1;
			while (itElements->hasNext()) {
				cout << id << ": " << itElements->next() << endl;
				id++;
			}
			cout << endl << endl;

			cout << "S:" << endl;
			itElements = s_dicts.at(*it)->listAll();
			id = 1;
			while (itElements->hasNext()) {
				cout << id << ": " << itElements->next() << endl;
				id++;
			}
			cout << endl << endl;

			cout << "O:" << endl;
			itElements = o_dicts.at(*it)->listAll();
			id = 1;
			while (itElements->hasNext()) {
				cout << id << ": " << itElements->next() << endl;
				id++;
			}
			cout << endl << endl;

			cout << "P:" << endl;
			itElements = p_dicts.at(*it)->listAll();
			id = 1;
			while (itElements->hasNext()) {
				cout << id << ": " << itElements->next() << endl;
				id++;
			}
			cout << endl << endl;
		}
*/
		/*
		 * END DICTIONARY TEST
		 */

		/*
		 //prepare the number of prior elements
		 vector<int> totalPriorSO(pow(2, numVersions), 0);
		 vector<int> totalPriorS(pow(2, numVersions), 0);
		 vector<int> totalPriorP(pow(2, numVersions), 0);
		 vector<int> totalPriorO(pow(2, numVersions), 0);


		 for (it = combinationsUniqTriples.begin(); //I have to do it for all the combinations of triples and not dictionaries because later I iterate by comb. of triples.
		 it != combinationsUniqTriples.end(); ++it) {
		 //cout << " COMBINATION:" << *it << endl;
		 //get the number of top level combinations
		 set<string> topLevels = getTopLevels(*it, numVersions);
		 //get the number of SO, S, P and O of these top level combinations

		 for (set<string>::iterator itLevels = topLevels.begin();
		 itLevels != topLevels.end(); ++itLevels) {
		 int currlevel = std::bitset<32>(*itLevels).to_ulong();
		 totalPriorSO.at(*it) +=
		 so_positionsPerCombination.at(currlevel).size();
		 totalPriorS.at(*it) +=
		 s_positionsPerCombination.at(currlevel).size();
		 totalPriorP.at(*it) +=
		 p_positionsPerCombination.at(currlevel).size();
		 totalPriorO.at(*it) +=
		 o_positionsPerCombination.at(currlevel).size();
		 }

		 //	cout<<"so_positionsPerCombination.at(*it).size():"<<so_positionsPerCombination.at(*it).size()<<endl;
		 //	cout<<"totalPriorSO.at(*it):"<<totalPriorSO.at(*it)<<endl;
		 }
		 */

		//cout << "Iterate triples per combination of triples" << endl;
		for (it = combinationsUniqTriples.begin();
				it != combinationsUniqTriples.end(); ++it) {

			/*
			 * prepare the number of prior elements
			 *
			 */

		//	cout << " COMBINATION:" << *it << endl;
			//get the number of top level combinations
			set<int, std::greater<int> > topLevels = getTopLevelsInt(*it,
					numVersions); //order in descending order
			//std::sort(topLevels.begin(), topLevels.end(), std::greater<int>());

			vector<int> totalPriorSO(pow(2, numVersions), 0);
			vector<int> totalPriorS(pow(2, numVersions), 0);
			vector<int> totalPriorP(pow(2, numVersions), 0);
			vector<int> totalPriorO(pow(2, numVersions), 0);

			//get the number of SO, S, P and O of these top level combinations

			//cout << "Iterating top levels (should be in order desc)" << endl;
			int prevSO = 0, prevS = 0, prevO = 0, prevP = 0;
			for (set<int>::iterator itLevels = topLevels.begin();
					itLevels != topLevels.end(); ++itLevels) {
				int currlevel = *itLevels;
				//cout << "current level:" << currlevel << endl;
				totalPriorSO.at(currlevel) = prevSO;
				prevSO += so_positionsPerCombination.at(currlevel).size();
				totalPriorS.at(currlevel) = prevS;
				prevS += s_positionsPerCombination.at(currlevel).size();
				totalPriorP.at(currlevel) = prevP;
				prevP += p_positionsPerCombination.at(currlevel).size();
				totalPriorO.at(currlevel) = prevO;
				prevO += o_positionsPerCombination.at(currlevel).size();
			}
			//insert current
			totalPriorSO.at(*it) = prevSO;
			totalPriorS.at(*it) = prevS;
			totalPriorO.at(*it) = prevO;
			totalPriorP.at(*it) = prevP;

			//test
			/*cout << "totalPriorSO:" << endl;
			for (int i = 0; i < totalPriorSO.size(); i++) {
				cout << i << ":" << totalPriorSO.at(i) << endl;
			}
			cout << endl;
			cout << "totalPriorS:" << endl;
			for (int i = 0; i < totalPriorS.size(); i++) {
				cout << i << ":" << totalPriorS.at(i) << endl;
			}
			cout << endl;
			cout << "totalPriorO:" << endl;
			for (int i = 0; i < totalPriorO.size(); i++) {
				cout << i << ":" << totalPriorO.at(i) << endl;
			}
			cout << endl;
			cout << "totalPriorP:" << endl;
			for (int i = 0; i < totalPriorP.size(); i++) {
				cout << i << ":" << totalPriorP.at(i) << endl;
			}
			cout << endl;
*/
			//	cout<<"so_positionsPerCombination.at(*it).size():"<<so_positionsPerCombination.at(*it).size()<<endl;
			//	cout<<"totalPriorSO.at(*it):"<<totalPriorSO.at(*it)<<endl;

			//	cout << " COMBINATION:" << *it << endl;

			for (int n = 0; n < triples_positionsPerCombination.at(*it).size();
					n++) {
				int currentPosition =
						triples_positionsPerCombination.at(*it).at(n);
				//cout << "currentPosition:" << currentPosition << endl;
				TripleID tid = *itPositionTriple->getTripleFromPosition(
						currentPosition);
				//cout << "insert into triplesListsperCombination" << endl;
				//insert into triplesListsperCombination
				triplesListsperCombination.at(*it)->insert(tid);
			}
			//	cout << "get the number of SO,S, O in this combination" << endl;
			//get the number of SO,S, O in this combination
			int totalPriorSOInThisCombination = prevSO
					+ so_positionsPerCombination.at(*it).size();
			int totalSInThisCombination = prevS
					+ s_positionsPerCombination.at(*it).size();
			int totalOInThisCombination = prevO
					+ o_positionsPerCombination.at(*it).size();

			ModifiableTriples* oldtriples = triplesListsperCombination.at(*it);
			ModifiableTriples* newtriples = new TriplesList(spec);
			IteratorTripleID *it2 = oldtriples->searchAll();
			while (it2->hasNext()) {
				TripleID *triple = it2->next();
				//cout << "old Triple: " << *triple << endl;
				//TODO REEMPLAZAR LOS IDS
				//cout<<*triple<<endl;

				string combinationElement = "";
				int position = triple->getSubject() - 1;
				for (int j = 0; j < numVersions; j++) {
					combinationElement = (
							bitmapsDictionariesGraph.at(j)->access(position) ?
									"1" : "0") + combinationElement;
				}
				int combinationStoringSubject = std::bitset<32>(
						combinationElement).to_ulong();

				//FIXME FALTARIA VER QUE ESTO LO HACE BIEN Y SOLUCIONAR GAPS DE LOS SUJETOS PARA LUEGO ESCRIBIR LOS TRIPLES
				string subjString = dictionary->idToString(triple->getSubject(),
						SUBJECT);
				if (triple->getSubject() <= dictionary->getNshared()) {
					int newSubject =
							so_dicts.at(combinationStoringSubject)->locate(
									(unsigned char*) subjString.c_str(),
									subjString.length());
					triple->setSubject(
							newSubject
									+ totalPriorSO.at(
											combinationStoringSubject));
				} else {
					int newSubject =
							s_dicts.at(combinationStoringSubject)->locate(
									(unsigned char*) subjString.c_str(),
									subjString.length());
					triple->setSubject(
							newSubject
									+ totalPriorS.at(combinationStoringSubject)
									+ totalPriorSOInThisCombination);
				}

				combinationElement = "";
				position = (triple->getPredicate()
						+ (dictionary->getNumberOfElements()
								- dictionary->getNpredicates()) - 1);
				for (int j = 0; j < numVersions; j++) {
					combinationElement = (
							bitmapsDictionariesGraph.at(j)->access(position) ?
									"1" : "0") + combinationElement;
				}
				int combinationStoringPredicate = std::bitset<32>(
						combinationElement).to_ulong();

				string predicateString = dictionary->idToString(
						triple->getPredicate(), PREDICATE);

				int newPredicate =
						p_dicts.at(combinationStoringPredicate)->locate(
								(unsigned char*) predicateString.c_str(),
								predicateString.length());
				triple->setPredicate(
						newPredicate
								+ totalPriorP.at(combinationStoringPredicate));

				combinationElement = "";
				if (triple->getObject() <= dictionary->getNshared()) {
					position = triple->getObject() - 1;
				} else {
					position = triple->getObject() + dictionary->getNsubjects()
							- dictionary->getNshared() - 1;
				}
				for (int j = 0; j < numVersions; j++) {
					combinationElement = (
							bitmapsDictionariesGraph.at(j)->access(position) ?
									"1" : "0") + combinationElement;
				}
				int combinationStoringObject = std::bitset<32>(
						combinationElement).to_ulong();

				string objString = dictionary->idToString(triple->getObject(),
						OBJECT);
				if (triple->getObject() <= dictionary->getNshared()) {
					int newObject =
							so_dicts.at(combinationStoringObject)->locate(
									(unsigned char*) objString.c_str(),
									objString.length());
					triple->setObject(
							newObject
									+ totalPriorSO.at(
											combinationStoringObject));
				} else {
					int newObject =
							o_dicts.at(combinationStoringObject)->locate(
									(unsigned char*) objString.c_str(),
									objString.length());
					//cout<<"newObject: "<<newObject<<endl;
					//cout<<"totalPriorSOInThisCombination: "<<totalPriorSOInThisCombination<<endl;
					//cout<<"totalPriorO.at(combinationStoringObject): "<<totalPriorO.at(combinationStoringObject)<<endl;
					triple->setObject(
							newObject + totalPriorO.at(combinationStoringObject)
									+ totalPriorSOInThisCombination);
				}

				newtriples->insert(*triple);
				//cout << "new Triple: " << *triple << endl << endl;
			}
			//FIXME UNCOMMENT TO SAVE THE TRIPLES

			//sort and remove duplicates (just in case)
			//newtriples->sort(SPO, NULL);
			// cout << "   Create Compact Triples (Not BT because of the gaps so far) and save Triples" << endl;
			// BitmapTriples* bt = new BitmapTriples(spec); //does not allow gaps
			//CompactTriples* ct = new CompactTriples(spec); //does not allow gaps
			PlainTriples* pt = new PlainTriples(spec);
			pt->load(*newtriples, NULL);
			ControlInformation ci;

			ci.clear();
			ci.setType(TRIPLES);

			if (singleFileOption) {
				pt->save(*out, ci, NULL);

			} else {
				ostream *outTriples;
				ofstream outFTriples;

				std::stringstream outfile;
				outfile << outputDirTriples << (*it) << ".bt";

				outFTriples.open(outfile.str().c_str());
				outTriples = &outFTriples;
				pt->save(*outTriples, ci, NULL);

			}
		}

		/*
		 cout << "test one combination, e.g. 991" << endl;
		 vector<int> pos991 = positionsPerCombination.at(991);
		 for (int p = 0; p < pos991.size(); p++) {
		 int id = pos991.at(p);
		 cout << "pos:" << id << endl;
		 if (id <= dictionary->getNshared()) {
		 cout << "   element subject/object:"
		 << dictionary->idToString(id, SUBJECT);
		 } else if (id <= dictionary->getNsubjects()) {
		 cout << "   element subject:"
		 << dictionary->idToString(id, SUBJECT);
		 } else if (id
		 > (dictionary->getNumberOfElements()
		 - dictionary->getNpredicates())) {
		 cout << "   element predicate:"
		 << dictionary->idToString(
		 dictionary->getNumberOfElements()-id,
		 PREDICATE);
		 } else {
		 cout << "   element object:"
		 << dictionary->idToString(
		 id - dictionary->getNsubjects()
		 + dictionary->getNshared(), OBJECT);
		 }
		 cout << endl;

		 }
		 */

		*outTime << st << endl;

		if(outputTimeFile!="") {
			outFTime.close();
		}

		delete hdt; // Remember to delete instance when no longer needed!
	} catch (char *e) {
		cout << "ERROR: " << e << endl;
	} catch (const char *e) {
		cout << "ERROR: " << e << endl;
	}
}

void ParseCallback::processTriple(hdt::TripleString& triple,
		unsigned long long pos) {
	//cout << "Triple String: " << triple << endl;
	count++;
	if ((count % 10000000) == 0)
		cout << count << " lines" << endl;

	string subject = triple.getSubject();
	string predicate = triple.getPredicate();
	string object = triple.getObject();
	//cout << "Result String: " << subject << ", " << predicate << ", " << object << endl;

	if (subject.at(0) == '<') {
		subject = subject.substr(1, subject.length() - 2);
	}
	if (object.at(0) == '<') {
		object = object.substr(1, object.length() - 2);
	}
	if (predicate.at(0) == '<') {
		predicate = predicate.substr(1, predicate.length() - 2);
	}
	//cout << "Result after filtering String: " << subject << ", " << predicate << ", " << object << endl;

	int subj = dictionary->stringToId(subject, SUBJECT);

	if (subj == 10) {
		cout << "subject is TEN!" << endl;
	}
	/*
	 * PUT SUBJECT IN THE BEGINING
	 */

	bitmapDictionariesGraph->set(subj - 1, true);

	//mark subject as present
	bitmapsSubjetsGraph->set(subj - 1, true);

	//cout << "Prueba Subject: " <<subj<<endl<<endl;
	int pred = dictionary->stringToId(predicate, PREDICATE);
	/*
	 * PUT at the end
	 */
	bitmapDictionariesGraph->set(
			(dictionary->getNumberOfElements() - dictionary->getNpredicates())
					+ pred - 1, true);

	//cout << "Prueba Predicate: " <<pred<<endl<<endl;
	int obj = dictionary->stringToId(object, OBJECT);
	//if (obj == 10) {
	//	cout << "obj is TEN!" << endl;
	//}
	//cout << "Prueba Object: " <<obj<<endl<<endl;
	/*
	 * PUT NON SHARED OBJECTS AFTER THE SUBJECTS
	 */
	if (obj <= dictionary->getNshared()) {
		bitmapDictionariesGraph->set(obj - 1, true);
	} else {
		bitmapDictionariesGraph->set(
				dictionary->getNsubjects() + obj - dictionary->getNshared() - 1,
				true);
	}

	//insert triple
	TripleID* tid = new TripleID(subj, pred, obj);

	//triplesListsGraph->insert(*tid);

	//get position of the triple
	int positionTriple = itgetPositionTriple->getNumberofTriple(*tid);
	//mark this position in the graphk
	bitmapTriplesGraph->set(positionTriple, true);
	if (subj == 0 || pred == 0 || obj == 0) {
		cerr << "Error:" << subject + " " + predicate + " " + object << endl;
		exit(-1);
	}
	/*if ((count % 10000) == 0)
	 cout << subj << " " << pred << " " << obj << " ." << endl;
	 */

}

